"""
The {{ cookiecutter.workchain_name }} workchain

----------------

{{ cookiecutter.workchain_docstring }}
"""
# pylint: disable=attribute-defined-outside-init
import random
import numpy as np

from aiida.common.extendeddicts import AttributeDict
from aiida.engine import calcfunction, WorkChain, while_, append_
from aiida.plugins import WorkflowFactory, DataFactory
from aiida_vasp.utils.workchains import prepare_process_inputs, compose_exit_code


class {{ cookiecutter.workchain_name[0]|upper}}{{cookiecutter.workchain_name[1:] }}WorkChain(WorkChain):
    """
    The {{ cookiecutter.workchain_name }} workchain
    
    ----------------

    {{ cookiecutter.workchain_docstring }}
    """

    _verbose = False
    _next_workchain_string = '{{ cookiecutter.next_workchain_to_call }}'
    _next_workchain = WorkflowFactory(_next_workchain_string)

    @classmethod
    def define(cls, spec):
        super({{ cookiecutter.workchain_name[0]|upper}}{{cookiecutter.workchain_name[1:] }}WorkChain, cls).define(spec)
        spec.expose_inputs(cls._next_workchain, exclude=['structure'])
        spec.input_namespace('structures',
                             valid_type=DataFactory('structure'),
                             dynamic=True,
                             help='a dictionary of structures to use')
        spec.exit_code(0, 'NO_ERROR', message='the sun is shining')
        spec.exit_code(420, 'ERROR_NO_CALLED_WORKCHAIN', message='no called workchain detected')
        spec.exit_code(500, 'ERROR_UNKNOWN', message='unknown error detected in the {{ cookiecutter.workchain_name }} workchain')

        spec.outline(
            cls.initialize,
            while_(cls.run_next_workchains)(
                cls.init_next_workchain,
                cls.run_next_workchain,
                cls.verify_next_workchain,
                cls.extract_quantities
            ),
            cls.finalize
        )  # yapf: disable

        spec.output('quantities',
                    valid_type=DataFactory('array'),
                    help='a container for various quantities')
        spec.expose_outputs(cls._next_workchain)

    def initialize(self):
        """Initialize the {{cookiecutter.workchain_name}} workchain."""
        self._init_context()
        self._init_inputs()

    def _init_context(self):
        """Initialize context variables that are used during the."""
        # Set the exit code to error in case we forget to set it to NO_ERROR
        self.ctx.exit_code = self.exit_codes.ERROR_UNKNOWN  # pylint: disable=no-member

        # Copy structures to context, since we will empty it as we go.
        # Since structures is an input to a workchain we cannot modify it and need to copy.
        self.ctx.structures = dict(self.inputs.structures)

        # Continue to submit workchains until this is True
        self.ctx.is_finished = False

        # Define an interation index
        self.ctx.interation = 0

        # Define the context inputs
        self.ctx.inputs = AttributeDict()

        # Define container to store quantities that is extracted in each step
        self.ctx.quantities_container = []

    def _init_inputs(self):
        """Initialize inputs."""

        # Set verbose flag (more or less details on the workchain report)
        try:
            self._verbose = self.inputs.verbose.value
        except AttributeError:
            pass

    def run_next_workchains(self):
        """
        Return whether a new workchain should be run.

        This is the case as long as the last workchain has not finished successfully.
        """
        return not self.ctx.is_finished

    def init_next_workchain(self):
        """Initialize the next workchain."""

        # Elevate iteration index
        self.ctx.iteration += 1

        # Check that the context inputs exists
        try:
            self.ctx.inputs
        except AttributeError:
            raise ValueError('No input dictionary was defined in self.ctx.inputs')

        # Take the exposed inputs and add them to the context input. This would
        # be the inputs you supply to this workchain
        self.ctx.inputs.update(self.exposed_inputs(self._next_workchain))

        # We did not expose the structure as we would like to set this
        # from the supplied structures input. Just choose any item in the
        # structures dictionary and asign that to the next run.
        item = random.choice(list(self.ctx.structures.keys()))
        self.ctx.inputs.structure = self.ctx.structures.pop(item)

        # Make sure we do not have any floating dict (convert to Dict etc.)
        self.ctx.inputs = prepare_process_inputs(self.ctx.inputs)

    def run_next_workchain(self):
        """
        Run the next workchain

        It is either submitted to the daemon or run, depending on how you
        run this workchain. The execution method is inherited.
        """
        inputs = self.ctx.inputs
        running = self.submit(self._next_workchain, **inputs)
        self.report('launching {}<{}> iteration #{}'.format(self._next_workchain.__name__, running.pk, self.ctx.iteration))
        self.to_context(workchains=append_(running))

    def verify_next_workchain(self):
        """Correct for unexpected behavior."""

        try:
            workchain = self.ctx.workchains[-1]
        except IndexError:
            self.report('There is no {} in the called workchain list.'.format(self._next_workchain.__name__))
            return self.exit_codes.ERROR_NO_CALLED_WORKCHAIN  # pylint: disable=no-member

        # Inherit exit status from last workchain (supposed to be
        # successfull)
        next_workchain_exit_status = workchain.exit_status
        next_workchain_exit_message = workchain.exit_message
        if not next_workchain_exit_status:
            self.ctx.exit_code = self.exit_codes.NO_ERROR  # pylint: disable=no-member
        else:
            self.ctx.exit_code = compose_exit_code(next_workchain_exit_status, next_workchain_exit_message)
            self.report('The called {}<{}> returned a non-zero exit status. '
                        'The exit status {} is inherited'.format(workchain.__class__.__name__, workchain.pk, self.ctx.exit_code))

        # Stop further executions of workchains if there are no more structure
        # entries in the structures dictionary
        if not self.ctx.structures:
            self.ctx.is_finished = True

        return self.ctx.exit_code

    def extract_quantities(self):
        """Extract the quantities you want and store them in the container."""

        # What happens in this routine needs modifications by you.
        # Typically you get the output of the called workchain by doing
        # workchain = self.ctx.workchains[-1]
        # some_output_quantity = workchain.outputs.some_output_quantity

        # An example which stores nonsense.
        self.ctx.quantities_container.append([self.ctx.iteration, 'Some quantity for iteration: {}'.format(self.ctx.iteration)])

    def finalize(self):
        """
        Finalize the workchain.

        Take the quantity container and set it as an output of this workchain.
        """
        # Due to data provenance we cannot return AiiDA data containers that have
        # not been passed through a calcfunction, workfunction or a workchain. Create this now.
        quantities_container = store_quantities(DataFactory('list')(list=self.ctx.quantities_container))

        # And then store the output on the workchain
        self.out('quantities', quantities_container)


@calcfunction
def store_quantities(quantities_container):
    """Stores the quantities to keep data provenance."""
    quantities_container_list = quantities_container.get_list()
    quantities_container_array = DataFactory('array')()
    quantities_container_array.set_array('quantities', np.array(quantities_container_list))
    return quantities_container_array



"""
Representation of CHGCAR objects.

Charge density data node (stores CHGCAR objects in the repository).
"""

# pylint: disable=abstract-method
# explanation: pylint wrongly complains about (aiida) Node not implementing query
from aiida.plugins import DataFactory

SinglefileData = DataFactory('core.singlefile')


class ChargedensityData(SinglefileData):  # pylint: disable=too-few-public-methods
    pass



"""
Module for converting between different representation for
initial site magnetization.
"""

import re


def create_additional_species(species: list[str], magmom: list[float]) -> tuple[list[str], dict[str, float]]:
    """
    Create additional species depending on magnetic moments.
    For example, create Fe1 and Fe2 if there are Fe with different
    magnetisations.

    Returns:
        a tuples of (newspecies, magmom_mapping)
    """

    unique_species = set(species)
    new_species = []
    current_species_mapping = {sym: {} for sym in unique_species}
    for symbol, this_mag in zip(species, magmom):
        current_symbol = symbol
        # Mappings for this original symbol
        mapping = current_species_mapping[symbol]
        # First check if this magmom has been treated
        not_seen = True
        for sym_, mag_ in mapping.items():
            if mag_ == this_mag:
                current_symbol = sym_
                not_seen = False
        # This symbol has not been seen yet
        if not_seen:
            if current_symbol in mapping:
                # The other species having the same symbol has been assigned
                counter = len(mapping) + 1
                current_symbol = f'{symbol}{counter}'
            mapping[current_symbol] = this_mag
        new_species.append(current_symbol)

    # Rename symbols that has more than one species, so A becomes A1
    for symbol, mapping in current_species_mapping.items():
        if len(mapping) > 1:
            mapping[f'{symbol}1'] = mapping[symbol]
            mapping.pop(symbol)
            # Refresh the new_species list
            new_species = [f'{sym}1' if sym == symbol else sym for sym in new_species]

    all_mapping: dict[str, float] = {}
    for value in current_species_mapping.values():
        all_mapping.update(value)

    return new_species, all_mapping


def convert_to_plain_list(species: list[str], magmom_mapping: dict[str, float]) -> tuple[list[str], list[float]]:
    """
    Covert from a decorated species list to a plain list of symbols
    and magnetic moments.

    Returns:
        A tuple of (symbols, magmoms)
    """
    magmoms: list[float] = []
    symbols: list[str] = []
    for symbol in species:
        magmoms.append(magmom_mapping[symbol])
        # Drop the number suffix in the symbol
        match = re.match(r'(\w+)\d+', symbol)
        if match:
            symbols.append(match.group(1))
        else:
            symbols.append(symbol)
    return symbols, magmoms



"""
VASP to Wannier90 calculation.

VASP2Wannier90 - Calculation.
"""

from typing import Any

from aiida import orm

# pylint: disable=abstract-method, unreachable, undefined-variable
# explanation: pylint wrongly complains about (aiida) Node not implementing query
from aiida.plugins import DataFactory

from aiida_vasp.calcs.vasp import VaspCalculation

# from aiida_wannier90.io import write_win  # pylint: disable=wrong-import-order, import-error


class Vasp2w90Calculation(VaspCalculation):
    """General purpose Calculation for using vasp with the vasp2wannier90 interface."""

    _default_parser = 'vasp.vasp2w90'

    @classmethod
    def define(cls, spec: Any) -> None:
        super(Vasp2w90Calculation, cls).define(spec)
        spec.input(
            'wannier_parameters',
            valid_type=orm.Dict,
            required=False,
            help='Input parameters for the Wannier90 interface.',
        )
        spec.input(
            'wannier_projections',
            valid_type=(orm.OrbitalData, orm.List),
            required=False,
            help='Projections to be defined in the Wannier90 input.',
        )

    def prepare_for_submission(self, folder: Any) -> None:
        """Override the method such that we can add the flag that executes Wannier90 in library mode."""
        raise NotImplementedError('The AiiDA-Wannier90 plugin is not yet v2 ready.')
        # Create a new parameters node that contain the flag that turns on the Wannier90 library mode such
        # that the Wannier90 inputs are created. Execution of Wannier90 itself is to be done after the execution
        # of this calculation with the aiida-wannier90 plugins calculation.
        parameters = self.inputs.parameters.get_dict()
        parameters.update({'lwannier90': True})
        self.inputs.parameters = DataFactory('core.dict')(dict=parameters)
        # Check if any Wannier90 parameters are given
        try:
            _ = self.inputs.wannier_parameters
        except AttributeError:
            self.inputs.wannier_parameters = DataFactory('core.dict')(dict={})

        # Then call the super function.
        return super().prepare_for_submission(folder)

    def write_win(self, dst: Any) -> None:
        """Write Wannier90 input."""
        raise NotImplementedError('The AiiDA-Wannier90 plugin is not yet v2 ready.')
        # write_win(filename=dst, parameters=self.inputs.wannier_parameters,
        # projections=self.inputs.wannier_projections)

    @staticmethod
    def new_wannier_parameters(**kwargs: Any) -> Any:
        return DataFactory('core.dict')(**kwargs)

    def write_additional(self, folder: Any, calcinfo: Any) -> None:
        super().write_additional(folder, calcinfo)
        win = folder.get_abs_path('wannier90.win')
        self.write_win(win)



"""
Test for the NEB workchain
"""

from io import StringIO

import pytest
from aiida import orm
from aiida.engine import run_get_node
from aiida.plugins import DataFactory, WorkflowFactory

from aiida_vasp.parsers.content_parsers.poscar import PoscarParser
from aiida_vasp.parsers.vasp import get_structure_node
from aiida_vasp.utils.mock_code import VaspMockRegistry
from aiida_vasp.utils.neb import neb_interpolate


@pytest.fixture
def nh3_end_points(fresh_aiida_env):
    """Making NH3 structure for NEB example"""
    pos1 = StringIO(
        """#NO COMMENT
  1.00000000000000
    6.000000    0.000000    0.000000
    0.000000    7.000000    0.000000
    0.000000    0.000000    8.000000
    H     N
    3     1
Direct
 0.636428  0.567457  0.5491645
 0.500000  0.364985  0.5491330
 0.363572  0.567457  0.5491645
 0.500000  0.500000  0.5000000
"""
    )
    pos2 = StringIO(
        """ammonia flipping
  1.00000000000000
    6.000000    0.000000    0.000000
    0.000000    7.000000    0.000000
    0.000000    0.000000    8.000000
    H     N
    3     1
Direct
 0.636428  0.567457  0.4508355
 0.500000  0.364985  0.4508670
 0.363572  0.567457  0.4508355
 0.500000  0.500000  0.5000000
"""
    )
    init = PoscarParser(handler=pos1).structure
    final = PoscarParser(handler=pos2).structure

    return get_structure_node(init), get_structure_node(final)


@pytest.fixture
def neb_wc_input(fresh_aiida_env, upload_potcar, potcar_family_name, potcar_mapping, mock_vasp_strict, nh3_end_points):
    """Generate inputs for an NEB workchain"""
    # upload_real_pseudopotentials('/home/bonan/appdir/VASP/POTCARS/potpaw_PBE.54-2015_subset/')
    init, final = nh3_end_points
    neb_frames = neb_interpolate(init, final, orm.Int(3))
    parameters = {
        'images': 3,
        'spring': -5,
        'ibrion': 3,
        'nsw': 50,
        'algo': 'normal',
        'potim': 0.0,
        'iopt': 1,
        'ediffg': -0.02,
    }
    kpoints = orm.KpointsData()
    kpoints.set_kpoints_mesh((1, 1, 1))
    builder = WorkflowFactory('vasp.neb').get_builder()
    builder.parameters = orm.Dict(dict={'incar': parameters})
    builder.options = orm.Dict(dict={'resources': {'tot_num_mpiprocs': 1, 'num_machines': 1}, 'withmpi': False})

    builder.potential_family = orm.Str(potcar_family_name)
    builder.potential_mapping = orm.Dict(dict=potcar_mapping)

    builder.kpoints = kpoints
    builder.initial_structure = neb_frames['image_init']
    builder.final_structure = neb_frames['image_final']
    builder.neb_images = {f'image_{i:02d}': neb_frames[f'image_{i:02d}'] for i in (1, 2, 3)}
    builder.code = mock_vasp_strict
    #    create_authinfo(computer=mock_vasp_strict.computer, store=True)

    builder.metadata.label = 'NH3 NEB'

    return builder


def upload_real_pseudopotentials(path):
    """
    Upload real pseudopotentials for workchain test mock deposition


    This function should be called once before the REAL vasp calculation is launch to setup the
    correct POTCARs
    """
    global POTCAR_FAMILY_NAME  # noqa: PLW0603
    POTCAR_FAMILY_NAME = 'TEMP'
    potcar_data_cls = DataFactory('vasp.potcar')
    potcar_data_cls.upload_potcar_family(path, 'TEMP', 'TEMP-REALPOTCARS', stop_if_existing=False, dry_run=False)


def upload_real_workchain(node, name):
    """
    Upload the workchain to the repository to make it work with mocking

    This function should be called once after the REAL vasp calculation is run during the test
    """

    reg = VaspMockRegistry()
    print(reg.base_path)
    reg.upload_aiida_work(node, name)


def test_vasp_neb_wc(fresh_aiida_env, neb_wc_input):
    """Test the workchain"""

    _, out_node = run_get_node(neb_wc_input)
    assert out_node.exit_status == 0
    assert 'image_01' in out_node.outputs.structure
    assert 'image_02' in out_node.outputs.structure
    assert 'image_03' in out_node.outputs.structure
    assert 'total_energies' in out_node.outputs.misc
    assert 'forces' in out_node.outputs.misc
    # upload_real_workchain(out_node, "neb-workchain")



"""
Module for settings up NEB calculations
"""

import os
from pathlib import Path
from typing import Any

from aiida import orm
from aiida.common import CalcInfo
from aiida.common.exceptions import InputValidationError
from aiida.common.folders import Folder

from aiida_vasp.calcs.vasp import VaspCalculation, ordered_unique_symbols
from aiida_vasp.data.chargedensity import ChargedensityData
from aiida_vasp.data.potcar import PotcarData
from aiida_vasp.data.wavefun import WavefunData
from aiida_vasp.parsers.content_parsers.poscar import PoscarParser


class VaspNEBCalculation(VaspCalculation):
    """
    NEB calculations using VASP

    Calculations for performing NEB calculations.
    NEB calculations requires standard VASP inputs, but POSCAR are placed in
    folder names 00, 01, 02... N for N-1 number of images.

    Input frames should be placed under the ``neb_images`` input namespace as a dictionary like::

      {
          'image_00': structure_1,
          'image_01': structure_2
          ....
      }

    Output of individual frames are placed in the corresponding namespace under the same convention.
    """

    # Use stdout as the name for the diverted stdout to be consistent with those in the image folders
    _VASP_OUTPUT = 'stdout'
    _ALWAYS_RETRIEVE_LIST = ['OUTCAR', 'vasprun.xml', _VASP_OUTPUT]
    _PER_IMAGE_ALWAYS_RETRIEVE_LIST = ['OUTCAR', 'CONTCAR', _VASP_OUTPUT]
    _query_type_string = 'vasp.neb'
    _plugin_type_string = 'vasp.neb'
    _default_parser = 'vasp.neb'

    @classmethod
    def define(cls, spec: Any) -> None:
        super(VaspNEBCalculation, cls).define(spec)
        # NEB calculation does not have the structure input port
        spec.inputs.pop('structure')

        # Define the inputs.
        # options is passed automatically.
        spec.input(
            'parameters',
            valid_type=orm.Dict,
            help='The VASP input parameters (INCAR).',
        )
        spec.input(
            'dynamics',
            valid_type=orm.Dict,
            help='The VASP parameters related to ionic dynamics, e.g. flags to set the selective dynamics',
            required=False,
        )
        spec.input(
            'initial_structure',
            valid_type=(orm.StructureData, orm.CifData),
            help='The input structure (POSCAR) for initial image.',
        )
        spec.input(
            'final_structure',
            valid_type=(orm.StructureData, orm.CifData),
            help='The input structure (POSCAR) for the final image.',
        )
        spec.input(
            'metadata.options.withmpi',
            valid_type=bool,
            default=True,
            help='Set the calculation to use mpi',
        )
        spec.input_namespace(
            'neb_images',
            valid_type=(orm.StructureData, orm.CifData),
            help='Starting structure for the NEB images',
            dynamic=True,
        )
        # Need namespace on this as it should also accept keys that are of `kind`. These are unknown
        # until execution.
        spec.input_namespace(
            'potential',
            valid_type=PotcarData,
            help='The potentials (POTCAR).',
            dynamic=True,
        )
        spec.input(
            'kpoints',
            valid_type=orm.KpointsData,
            help='The kpoints to use (KPOINTS).',
        )
        spec.input_namespace(
            'charge_density',
            dynamic=True,
            valid_type=ChargedensityData,
            required=False,
            help='The charge density. (CHGCAR)',
        )
        spec.input_namespace(
            'wavefunctions',
            valid_type=WavefunData,
            dynamic=True,
            required=False,
            help='The wave function coefficients. (WAVECAR)',
        )
        spec.input(
            'settings',
            valid_type=orm.Dict,
            required=False,
            help='Additional parameters not related to VASP itself.',
        )
        spec.input(
            'metadata.options.parser_name',
            default=cls._default_parser,
        )

        # Define outputs.
        # remote_folder and retrieved are passed automatically

        # Turn on dynamic outputs to avoid problem when catching is used
        # it appears the catching implementation does not work with calcjobs with namespace outputs
        spec.outputs.dynamic = True
        spec.output_namespace(
            'structure',
            required=True,
            valid_type=orm.StructureData,
            help='NEB images',
            dynamic=True,
        )
        spec.output(
            'misc',
            valid_type=orm.Dict,
            required=True,
            help='Per-image misc output.',
        )
        spec.output(
            'arrays',
            valid_type=orm.ArrayData,
            required=False,
            help='Array data from each image.',
        )
        spec.output_namespace(
            'trajectory',
            valid_type=orm.TrajectoryData,
            dynamic=True,
            required=False,
            help='The NEB trajectory.',
        )

    def prepare_for_submission(self, folder: Folder) -> CalcInfo:
        """
        Add all files to the list of files to be retrieved.

        Notice that we here utilize both the retrieve batch of files, which are always stored after retrieval and
        the temporary retrieve list which is automatically cleared after parsing.
        """
        calcinfo = super().prepare_for_submission(folder)

        nimages = len(self.inputs.neb_images)
        nimage_keys = sorted(list(self.inputs.neb_images.keys()))

        image_folders = []

        # Iterate though each image_folder that needs to be setup
        for i in range(nimages + 2):
            folder_id = f'{i:02d}'
            image_folder = Path(folder.get_abs_path(folder_id))
            image_folder.mkdir()
            poscar = str(image_folder / 'POSCAR')
            if i == 0:
                # Write the initial image
                self.write_neb_poscar(self.inputs.initial_structure, poscar)
            elif i == nimages + 1:
                # Write the final image
                self.write_neb_poscar(self.inputs.final_structure, poscar)
            else:
                # Write NEB images
                img_key = nimage_keys[i - 1]
                self.write_neb_poscar(self.inputs.neb_images[img_key], poscar)
                image_folders.append(folder_id)

                # Link with singlefile WAVECAR/CHGCAR is needed
                if self._need_wavecar():
                    wavecar = self.inputs.wave_functions[img_key]
                    dst = folder_id + '/' + 'WAVECAR'
                    calcinfo.local_copy_list.append((wavecar.uuid, wavecar.filename, dst))

                if self._need_chgcar():
                    chgcar = self.inputs.charge_densities[img_key]
                    dst = folder_id + '/' + 'CHGCAR'
                    calcinfo.local_copy_list.append((chgcar.uuid, chgcar.filename, dst))
        try:
            store = self.inputs.settings.base.attributes.get('ALWAYS_STORE', default=True)
        except AttributeError:
            store = True

        try:
            additional_retrieve_list = self.inputs.settings.base.attributes.get(
                'PER_IMAGE_ADDITIONAL_RETRIEVE_LIST', default=[]
            )
        except AttributeError:
            additional_retrieve_list = []
        try:
            additional_retrieve_temp_list = self.inputs.settings.base.attributes.get(
                'PER_IMAGE_ADDITIONAL_RETRIEVE_TEMPORARY_LIST', default=[]
            )
        except AttributeError:
            additional_retrieve_temp_list = []

        if store:
            calcinfo.retrieve_list.extend(
                image_folder_paths(image_folders, set(self._PER_IMAGE_ALWAYS_RETRIEVE_LIST + additional_retrieve_list))
            )
            calcinfo.retrieve_temporary_list.extend(image_folder_paths(image_folders, additional_retrieve_temp_list))
        else:
            calcinfo.retrieve_temporary_list.extend(
                image_folder_paths(
                    image_folders, set(self._PER_IMAGE_ALWAYS_RETRIEVE_LIST + additional_retrieve_temp_list)
                )
            )
            calcinfo.retrieve_list.extend(image_folder_paths(image_folders, additional_retrieve_list))

        # self.logger.warning('Calcinfo: {}'.format(calcinfo))

        return calcinfo

    def _structure(self) -> orm.StructureData:
        """
        Get the input structure as AiiDa StructureData.

        This is required in order to support CifData as input as well.
        """
        structure = self.inputs.initial_structure
        if not hasattr(structure, 'get_pymatgen'):
            structure = orm.StructureData(ase=structure.get_ase())
        return structure

    def remote_copy_restart_folder(self) -> list[tuple[str, str, str]]:
        """
        Add all files required for restart to the list of files to be copied from the previous calculation.

        For NEB calculations, the CHGCAR and WAVECAR needs to be copied for each of the actual images.
        """
        restart_folder = self.inputs.restart_folder
        computer = self.node.computer
        included = ['CHGCAR', 'WAVECAR']
        nimages = len(self.inputs.neb_images)
        copy_list = []
        for image_id in range(1, nimages + 1):
            fdname = f'{image_id:02d}'

            existing_files = restart_folder.listdir(fdname)
            for name in included:
                if name not in existing_files:
                    # Here we simple issue an warning as the requirement of files will be explicitly checked by
                    # `write_additional` method
                    self.report(f'WARNING: File {name} does not exist in the restart folder.')
                else:
                    copy_list.append(
                        (
                            computer.uuid,
                            os.path.join(restart_folder.get_remote_path(), fdname, name),
                            os.path.join(fdname, name),
                        )
                    )

        return copy_list

    def write_neb_poscar(
        self, structure: orm.StructureData | orm.CifData, dst: str, positions_dof: list | None = None
    ) -> None:
        """
        Write the POSCAR.

        Passes the structures node (StructureData) to the POSCAR parser for
        preparation and writes to dst.

        :param dst: absolute path of the file to write to

        """
        settings = self.inputs.get('settings')
        settings = settings.get_dict() if settings else {}
        poscar_precision = settings.get('poscar_precision', 10)
        if positions_dof is not None:
            options = {'positions_dof': positions_dof}
        else:
            options = None
        poscar_parser = PoscarParser(data=ensure_structure_data(structure), precision=poscar_precision, options=options)
        poscar_parser.write(dst)

    def verify_inputs(self) -> None:
        """
        Verify the order of elements
        """
        super().verify_inputs()
        last_order = None
        last_num_sites = None
        for structure in list(self.inputs.neb_images.values()) + [
            self.inputs.initial_structure,
            self.inputs.final_structure,
        ]:
            # Convert to StructureData from CifData on demand....
            if not hasattr(structure, 'get_pymatgen'):
                structure_data = orm.StructureData(ase=structure.get_ase())
            else:
                structure_data = structure

            num_sites = len(structure_data.sites)

            order = ordered_unique_symbols(structure_data)
            if last_order is not None and order != last_order:
                raise InputValidationError('The input structures have non-euqal element orders.')
            if last_order is not None and num_sites != last_num_sites:
                raise InputValidationError('The input structures have non-equal number of atoms.')
            last_order = order
            last_num_sites = num_sites

    def write_incar(self, dst: str, validate_tags: bool = False) -> None:
        """
        Write the INCAR without tag validation.
        Validation is performed at `parsevasp` level and VTST tags are not included.

        :param dst: absolute path of the file to write to
        """
        super().write_incar(dst, validate_tags=validate_tags)


def image_folder_paths(image_folders: list[str], retrieve_names: list[str]) -> list[list[str | int]]:
    """
    Return a list of folders paths to be retrieved
    """
    retrieve_list = []
    for key in retrieve_names:
        for fdname in image_folders:
            # Skip the stdout for the first image - it is diverted to the root folder
            if int(fdname) == 1 and key == VaspNEBCalculation._VASP_OUTPUT:  # pylint: disable=protected-access
                continue
            # Need to use the tuple format to keep the sub directory structure
            retrieve_list.append([fdname + '/' + key, '.', 2])
    return retrieve_list


def ensure_structure_data(structure: orm.StructureData | orm.CifData) -> orm.StructureData:
    """
        Get the input structure as AiiDA StructureData.

    This is required in order to support CifData as input as well.
    """
    if not hasattr(structure, 'get_pymatgen'):
        structure = orm.StructureData(ase=structure.get_ase())
    return structure



"""Unit tests for aiida_vasp.calcs.base."""

# pylint: disable=unused-import,unused-argument,redefined-outer-name, import-outside-toplevel
import pytest
from aiida.common.extendeddicts import AttributeDict
from aiida.engine.utils import instantiate_process
from aiida.manage.manager import get_manager

from aiida_vasp.calcs.base import VaspCalcBase


def test_generate_base_calc(base_calc):
    """Test that it is possible to start the generation of an instance of the base calculation class."""

    manager = get_manager()
    runner = manager.get_runner()
    inputs = AttributeDict()
    metadata = AttributeDict({'options': {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}})
    inputs.metadata = metadata

    # Need more input, so we will get an AttributeError because we do not pass code
    with pytest.raises(AttributeError):
        instantiate_process(runner, VaspCalcBase, **inputs)

    # Let us try again with a code attached, thus should pass
    inputs.code = base_calc.inputs.code
    instantiate_process(runner, VaspCalcBase, **inputs)



"""
This module contains the class that prepares a specific VASP calculation.
"""

from __future__ import annotations

# encoding: utf-8
# pylint: disable=abstract-method
# explanation: pylint wrongly complains about (aiida) Node not implementing query
import os
from typing import TYPE_CHECKING, Any

from aiida import orm
from aiida.common.exceptions import InputValidationError, ValidationError
from aiida.engine import ProcessBuilder
from aiida.orm.nodes.data.base import to_aiida_type

from aiida_vasp.calcs.base import VaspCalcBase
from aiida_vasp.data.chargedensity import ChargedensityData
from aiida_vasp.data.potcar import PotcarData
from aiida_vasp.data.wavefun import WavefunData
from aiida_vasp.parsers.content_parsers.incar import IncarParser
from aiida_vasp.parsers.content_parsers.kpoints import KpointsParser
from aiida_vasp.parsers.content_parsers.poscar import PoscarParser
from aiida_vasp.parsers.content_parsers.potcar import MultiPotcarIo

if TYPE_CHECKING:
    from aiida.common import CalcInfo
    from aiida.common.folders import Folder


class VaspCalculation(VaspCalcBase):
    """
    General-purpose VASP calculation.

    By default retrieves only the 'OUTCAR', 'vasprun.xml', 'CONTCAR'
    and Wannier90 input / output objects. These objects are deleted after parsing.
    Additional retrieve objects can be specified via the
    ``settings['ADDITIONAL_RETRIEVE_TEMPORARY_LIST']`` input. In addition, if you want to keep
    any objects after parsing, put them in ``settings['ADDITIONAL_RETRIEVE_LIST']`` which is empty
    by default.

    Floating point precision for writing POSCAR objects can be adjusted using
    ``settings['poscar_precision']``, default: 10

    The following assumes you are familiar with the AiiDA data structures and
    how to set up and run an AiiDA calculation in general.

    Example usage::

        from aiida.orm import CalculationFactory, DataFactory
        from aiida.work import submit

        proc = CalculationFactory('vasp.vasp').process()
        inputs = proc.get_inputs_template()
        inputs.parameter = <Dict with INCAR params>
        inputs.structure = <StructureData>
        inputs.kpoints = <KpointsData>
        inputs.settings = <Dict with parser settings etc.>
        inputs.potential = DataFactory('vasp.potcar').get_potcars_from_structure(structure, ...)
        inputs.code = <Code representing vasp on your cluster>

        submit(proc, **inputs)

    Which is very similar to the workchain example.

    Since we do not want the content parsers to know about the AiiDA infrastructure,
    i.e. processes etc. we have no access to the exit codes defined on the CalcJob.
    We thus have to deal with failures in parsing directly in the write calls here.

    """

    _VASP_OUTPUT = 'vasp_output'
    _ALWAYS_RETRIEVE_LIST = ['CONTCAR', 'OUTCAR', 'vasprun.xml', 'wannier90*', _VASP_OUTPUT]
    _query_type_string = 'vasp.vasp'
    _plugin_type_string = 'vasp.vasp'

    @classmethod
    def define(cls, spec: Any) -> None:
        super(VaspCalculation, cls).define(spec)
        # Define the inputs.
        # options is passed automatically.
        spec.input(
            'parameters',
            valid_type=orm.Dict,
            serializer=to_aiida_type,
            help='The VASP input parameters (INCAR).',
        )
        spec.input(
            'dynamics',
            valid_type=orm.Dict,
            serializer=to_aiida_type,
            help='The VASP parameters related to ionic dynamics, e.g. flags to set the selective dynamics',
            required=False,
        )
        spec.input(
            'structure',
            valid_type=(orm.StructureData, orm.CifData),
            help='The input structure (POSCAR).',
        )

        # Need namespace on this as it should also accept keys that are of `kind`. These are unknown
        # until execution.
        spec.input_namespace(
            'potential',
            valid_type=PotcarData,
            help='The potentials (POTCAR).',
            dynamic=True,
        )
        spec.input(
            'kpoints',
            valid_type=orm.KpointsData,
            help='The kpoints to use (KPOINTS).',
        )
        spec.input(
            'charge_density',
            valid_type=ChargedensityData,
            required=False,
            help='The charge density. (CHGCAR)',
        )
        spec.input(
            'wavefunctions',
            valid_type=WavefunData,
            required=False,
            help='The wave function coefficients. (WAVECAR)',
        )
        spec.input(
            'settings',
            valid_type=orm.Dict,
            serializer=to_aiida_type,
            required=False,
            help='Additional parameters not related to VASP itself.',
        )
        spec.input(
            'metadata.options.parser_name',
            default='vasp.vasp',
        )
        spec.input(
            'vdw_kernel',
            valid_type=orm.SinglefileData,
            required=False,
            help='The vdw_kerenl.bindat file to be used for vdw calculations.',
        )

        # Define outputs.
        # remote_folder and retrieved are passed automatically
        spec.output(
            'misc',
            valid_type=orm.Dict,
            help='The output parameters containing smaller quantities that do not depend on system size.',
        )
        # Mark misc as the default output node
        spec.default_output_node = 'misc'
        spec.output(
            'structure',
            valid_type=orm.StructureData,
            required=False,
            help='The output structure.',
        )
        spec.output(
            'kpoints',
            valid_type=orm.KpointsData,
            required=False,
            help='The output k-points.',
        )
        spec.output(
            'trajectory',
            valid_type=orm.TrajectoryData,
            required=False,
            help='The output trajectory data.',
        )
        spec.output(
            'arrays',
            valid_type=orm.ArrayData,
            required=False,
            help='The output trajectory data.',
        )
        spec.output(
            'chgcar',
            valid_type=ChargedensityData,
            required=False,
            help='The output charge density CHGCAR file.',
        )
        spec.output(
            'wavecar',
            valid_type=WavefunData,
            required=False,
            help='The output plane wave coefficients file.',
        )
        spec.output(
            'bands',
            valid_type=orm.BandsData,
            required=False,
            help='The output band structure.',
        )
        spec.output(
            'dos',
            valid_type=orm.ArrayData,
            required=False,
            help='The output dos.',
        )
        spec.output(
            'energies',
            valid_type=orm.ArrayData,
            required=False,
            help='Energies of the calculation at each ionic/electronic step.',
        )
        spec.output(
            'projectors',
            valid_type=orm.ArrayData,
            required=False,
            help='The projectors for the calculation.',
        )
        spec.output(
            'parameters', valid_type=orm.Dict, required=False, help='All input parameters including the default values.'
        )
        # Standalone array quantities
        for name in ['hessian', 'dynmat', 'born_charges', 'dielectrics']:
            spec.output(
                name,
                valid_type=orm.ArrayData,
                required=False,
                help='The output {name} data.',
            )
        spec.exit_code(0, 'NO_ERROR', message='the sun is shining')
        spec.exit_code(
            350,
            'ERROR_NO_RETRIEVED_FOLDER',
            message='the retrieved folder data node could not be accessed.',
            invalidates_cache=True,
        )
        spec.exit_code(
            351,
            'ERROR_NO_RETRIEVED_TEMPORARY_FOLDER',
            message='the retrieved_temporary folder data node could not be accessed.',
            invalidates_cache=True,
        )
        spec.exit_code(
            352,
            'ERROR_CRITICAL_MISSING_OBJECT',
            message='an object that is marked by the parser as critical is missing.',
            invalidates_cache=True,
        )
        spec.exit_code(
            333,
            'ERROR_VASP_DID_NOT_EXECUTE',
            message='VASP did not produce any output and did likely not execute properly.',
            invalidates_cache=True,
        )

        # 700 series of the errors catches VASP execution related problems
        spec.exit_code(
            700,
            'ERROR_DID_NOT_FINISH',
            message='Calculation did not reach the end of execution.',
            invalidates_cache=True,
        )
        spec.exit_code(
            701,
            'ERROR_ELECTRONIC_NOT_CONVERGED',
            message='The electronic structure is not converged.',
        )
        spec.exit_code(
            702,
            'ERROR_IONIC_NOT_CONVERGED',
            message='The ionic relaxation is not converged.',
        )
        spec.exit_code(
            703,
            'ERROR_VASP_CRITICAL_ERROR',
            message='VASP calculation encountered a critical error: {error_message}.',
        )
        spec.exit_code(
            704,
            'ERROR_DIAGNOSIS_OUTPUTS_MISSING',
            message='Outputs for diagnosis are missing, please make sure `run_status` and `notifications`'
            ' quantities are requested for parsing.',
        )

        spec.exit_code(
            1002,
            'ERROR_NOT_ABLE_TO_PARSE_QUANTITY',
            message='the parser is not able to parse the {quantity} quantity which is marked as required',
        )
        spec.exit_code(
            1004,
            'ERROR_NOT_ABLE_TO_CREATE_NODE',
            message='the parser is not able to compose one or more output nodes: {nodes}',
        )
        spec.exit_code(1005, 'ERROR_OVERFLOW_IN_XML', message='Overflow detected in XML while parsing.')

    def prepare_for_submission(self, folder: Folder) -> CalcInfo:
        """
        Add all objects to the list of objects to be retrieved.

        Notice that we here utilize both the retrieve batch of objects, which are always stored after retrieval and
        the temporary retrieve list which is automatically cleared after parsing.
        """
        calcinfo = super().prepare_for_submission(folder)

        # Combine stdout and stderr into vasp_output so that the stream parser can parse it later.
        calcinfo.codes_info[0].stdout_name = self._VASP_OUTPUT
        calcinfo.codes_info[0].join_files = True

        # Still need the exceptions in case settings is not defined on inputs
        # Check if we want to store all always retrieve objects
        try:
            store = self.inputs.settings.base.attributes.get('ALWAYS_STORE', default=True)
        except AttributeError:
            store = True
        try:
            additional_retrieve_list = self.inputs.settings.base.attributes.get('ADDITIONAL_RETRIEVE_LIST', default=[])
        except AttributeError:
            additional_retrieve_list = []
        try:
            additional_retrieve_temp_list = self.inputs.settings.base.attributes.get(
                'ADDITIONAL_RETRIEVE_TEMPORARY_LIST', default=[]
            )  # pylint: disable=invalid-name
        except AttributeError:
            additional_retrieve_temp_list = []
        if store:
            calcinfo.retrieve_list = list(set(self._ALWAYS_RETRIEVE_LIST + additional_retrieve_list))
            calcinfo.retrieve_temporary_list = additional_retrieve_temp_list  # pylint: disable=invalid-name
        else:
            calcinfo.retrieve_temporary_list = list(set(self._ALWAYS_RETRIEVE_LIST + additional_retrieve_temp_list))
            calcinfo.retrieve_list = additional_retrieve_list
        try:
            provenance_exclude_list = self.inputs.settings.base.attributes.get('PROVENANCE_EXCLUDE_LIST', default=[])
        except AttributeError:
            provenance_exclude_list = []
        # Always include POTCAR in the exclude list (not added to the repository, regardless of store)
        calcinfo.provenance_exclude_list = list(set(provenance_exclude_list + ['POTCAR']))

        return calcinfo

    def verify_inputs(self) -> None:
        super().verify_inputs()
        _parameters = self.inputs.parameters.get_dict()
        _lorbit = _parameters.get('lorbit', 0)
        if 'settings' in self.inputs:
            _settings = self.inputs.settings.get_dict()
        else:
            _settings = {}
        _site_magnetization = 'site_magnetization' in _settings.get('parser_settings', {}).get('include_quantity', [])
        if _site_magnetization and _lorbit < 10:
            raise InputValidationError(f'Site magnetization requires "LORBIT>=10", value given {_lorbit}')

    @property
    def _parameters(self) -> dict[str, Any]:
        """Make sure all parameters are lowercase."""
        all_parameters = self.inputs.parameters.get_dict()
        try:
            return {k.lower(): v for k, v in all_parameters.items()}
        except KeyError:
            return {}

    def _need_kp(self) -> bool:
        """
        Return wether an input kpoints node is needed or not.

        :return output: True if input kpoints node is needed False otherwise needs 'parameters' input to be set.
        """
        return not bool('kspacing' in self._parameters or 'kgamma' in self._parameters)

    def _need_chgcar(self) -> bool:
        """
        Test wether an charge_densities input is needed or not.

        :return output: True if CHGCAR must be present False otherwise.

        """
        ichrg_d = 0 if self._need_wavecar() else 2
        icharg = self._parameters.get('icharg', ichrg_d)
        return bool(icharg in [1, 11])

    def _need_wavecar(self) -> bool:
        """
        Test wether a wavefunctions input is needed or not.

        :return output: True if WAVECAR must be present used False otherwise.
        """
        istrt_d = 1 if self.inputs.get('wavefunctions') else 0
        istart = self._parameters.get('istart', istrt_d)
        return bool(istart in [1, 2, 3])

    def _structure(self) -> orm.StructureData:
        """
        Get the input structure as AiiDa StructureData.

        This is required in order to support CifData as input as well.
        """
        structure = self.inputs.structure
        if not hasattr(structure, 'get_pymatgen'):
            structure = orm.StructureData(ase=structure.get_ase())
        return structure

    def write_additional(self, folder: Folder, calcinfo: CalcInfo) -> None:
        """Write CHGAR and WAVECAR if needed."""
        super().write_additional(folder, calcinfo)
        # a list of object names to be copied
        remote_copy_fnames = [os.path.split(entry[1])[1] for entry in calcinfo.remote_copy_list]
        if self._need_chgcar():
            # If we restart, we do not require inputs, but we should have a basic check
            # that CHGCAR is present
            if not self._is_restart():
                self.write_chgcar('CHGCAR', calcinfo)
            else:
                remote_folder = self.inputs.restart_folder
                if 'CHGCAR' not in remote_copy_fnames:
                    raise FileNotFoundError(f'Could not find CHGCAR in {remote_folder.get_remote_path()}')
        if self._need_wavecar():
            # If we restart, we do not require inputs, but we should have a basic check
            # that WAVECAR is present
            if not self._is_restart():
                self.write_wavecar('WAVECAR', calcinfo)
            else:
                remote_folder = self.inputs.restart_folder
                if 'WAVECAR' not in remote_copy_fnames:
                    raise FileNotFoundError(f'Could not find WAVECAR in {remote_folder.get_remote_path()}')

        # Process the vdw_kernel input
        if 'vdw_kernel' in self.inputs:
            calcinfo.local_copy_list.append(
                (self.inputs.vdw_kernel.uuid, self.inputs.vdw_kernel.filename, 'vdw_kernel.bindat')
            )

    def write_incar(self, dst: str, validate_tags: bool = True) -> None:
        """
        Write the INCAR.

        Passes the parameters node (Dict) from to the INCAR parser for
        preparation and writes to dst.

        :param dst: absolute path of the object to write to

        """
        # Check if parameters validation is turned off
        if self.inputs.get('settings'):
            settings_dict = self.inputs.settings.get_dict()
            if settings_dict.get('skip_parameters_validation') or settings_dict.get('unsupported_parameters'):
                validate_tags = False
        try:
            incar_parser = IncarParser(data=self.inputs.parameters, validate_tags=validate_tags)
            incar_parser.write(dst)
        except SystemExit as parser_error:
            raise ValidationError('The INCAR content did not pass validation.') from parser_error

    def write_poscar(self, dst: str) -> None:
        """
        Write the POSCAR.

        Passes the structures node (StructureData) to the POSCAR parser for
        preparation and writes to dst.

        :param dst: absolute path of the object to write to
        """
        settings = self.inputs.get('settings')
        settings = settings.get_dict() if settings else {}
        poscar_precision = settings.get('poscar_precision', 10)
        dynamics = self.inputs.get('dynamics')
        options = None
        if dynamics is not None:
            dynamics = dynamics.get_dict()
            positions_dof = dynamics.get('positions_dof')
            if positions_dof is not None:
                options = {'positions_dof': positions_dof}

        try:
            poscar_parser = PoscarParser(data=self._structure(), precision=poscar_precision, options=options)
            poscar_parser.write(dst)
        except SystemExit as parser_error:
            raise ValidationError('The POSCAR content did not pass validation.') from parser_error

    def write_potcar(self, dst: str) -> None:
        """
        Concatenates multiple POTCARs into one in the same order as the elements appear in POSCAR.

        :param dst: absolute path of the object to write to
        """
        structure = self._structure()
        multi_potcar = MultiPotcarIo.from_structure(structure, self.inputs.potential)
        multi_potcar.write(dst)

    def write_kpoints(self, dst: str) -> None:
        """
        Write the KPOINTS.

        Passes the kpoints node (KpointsData) to the KPOINTS parser for
        preparation and writes to dst.

        :param dst: absolute path of the object to write to
        """
        try:
            kpoint_parser = KpointsParser(data=self.inputs.kpoints)
            kpoint_parser.write(dst)
        except SystemExit as parser_error:
            raise ValidationError('The KPOINTS content did not pass validation.') from parser_error

    def write_chgcar(self, dst: str, calcinfo: CalcInfo) -> None:
        charge_density = self.inputs.charge_density
        calcinfo.local_copy_list.append((charge_density.uuid, charge_density.filename, dst))

    def write_wavecar(self, dst: str, calcinfo: CalcInfo) -> None:
        wave_functions = self.inputs.wavefunctions
        calcinfo.local_copy_list.append((wave_functions.uuid, wave_functions.filename, dst))

    @classmethod
    def immigrant(cls, code: orm.Code, remote_path: str, **kwargs: Any) -> tuple[type, ProcessBuilder]:
        """
        Returns VaspImmigrant class and associated inputs. This method will be obsolete at v3.0

        :param code: a Code instance for the code originally used.
        :param remote_path: The directory on the code's computer in which the simulation was run.
        :param metadata: dict.
        :param settings: dict. Used for non-default parsing instructions, etc.
        :param potential_family: str.
        :param potential_mapping: dict.
        :param use_wavecar: bool. Try to read WAVECAR.
        :param use_chgcar bool. Try to read CHGCAR.

        """
        from aiida_vasp.calcs.immigrant import VaspImmigrant  # noqa: PLC0415

        proc_cls = VaspImmigrant
        builder = proc_cls.get_builder_from_folder(code, str(remote_path), **kwargs)
        options = {'max_wallclock_seconds': 1, 'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}
        builder.metadata = kwargs.get('metadata', {'options': options})
        options = builder.metadata.get('options', options)
        max_wallclock_seconds = options.get('max_wallclock_seconds', 1)
        resources = options.get('resources', {'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        builder.metadata['options']['max_wallclock_seconds'] = max_wallclock_seconds  # pylint: disable=no-member
        builder.metadata['options']['resources'] = resources  # pylint: disable=no-member
        settings = kwargs.get('settings', {})
        settings.update({'import_from_path': str(remote_path)})
        builder.settings = orm.Dict(dict=settings)

        return proc_cls, builder


def ordered_unique_list(in_list: list) -> list:
    """List unique elements in input list, in order of first occurrence."""
    out_list = []
    for i in in_list:
        if i not in out_list:
            out_list.append(i)
    return out_list


def ordered_unique_symbols(structure: orm.StructureData) -> list[str]:
    """
    Return a list of ordered unique symbols in the structure
    """
    symbols = [structure.get_kind(kindname).symbol for kindname in structure.get_site_kindnames()]
    return ordered_unique_list(symbols)



import pathlib

import numpy as np
import pytest
from aiida import orm

from aiida_vasp.parsers.vasp import NotificationComposer, ParserSettingsConfig, VaspParser, is_all_empty


def test_all_empty_check():
    """Test the is_all_empty function"""
    empty = {'A': [], 'B': {'C': {}}}
    assert is_all_empty(empty)
    empty = []
    assert is_all_empty(empty)
    empty = {'A': []}
    assert is_all_empty(empty)
    not_empty = {'A': {'B': [1]}}
    assert not is_all_empty(not_empty)
    not_empty = {'A': [1, 2], 'B': 2}
    assert not is_all_empty(not_empty)
    assert not is_all_empty([1])


def test_parser_bare(calc_with_retrieved, request):
    _relative_file_path = '../test_data/basic_run'
    file_path = str(pathlib.Path(request.fspath).parent / _relative_file_path)
    node = calc_with_retrieved(file_path, {})
    parser = VaspParser(node)
    parser.parse(retrieved_tempoary_folder=file_path)
    assert 'misc' in parser.outputs
    assert 'site_magnetization' not in parser.outputs.misc.get_dict()
    assert 'structure' in parser.outputs

    node = calc_with_retrieved(file_path, {'parser_settings': {'include_quantity': ['projectors']}})
    parser = VaspParser(node)
    parser.parse(retrieved_tempoary_folder=file_path)
    assert 'dielectrics`' not in parser.outputs
    assert 'born_charges' not in parser.outputs
    assert 'parameters' not in parser.outputs['misc']
    assert 'trajectory' not in parser.outputs
    assert 'energies' not in parser.outputs

    # Test the parameters outputs
    node = calc_with_retrieved(file_path, {'parser_settings': {'include_quantity': ['parameters']}})
    parser = VaspParser(node)
    parser.parse(retrieved_tempoary_folder=file_path)
    assert 'parameters' in parser.outputs['misc']
    assert 'system' in parser.outputs['misc']['parameters']


@pytest.fixture
def parser_with_retrieved(calc_with_retrieved, request):
    """Fixture to create a VaspParser instance with a given pre-computed data folder"""

    def wrapped(name, settings={}, parse=True):
        _relative_file_path = f'../test_data/{name}'
        file_path = str(pathlib.Path(request.fspath).parent / _relative_file_path)
        node = calc_with_retrieved(file_path, settings)
        parser = VaspParser(node)
        exit_code = None
        if parse:
            exit_code = parser.parse(retrieved_tempoary_folder=file_path)
        return parser, exit_code

    return wrapped


@pytest.fixture
def parser_with_vasprun(parser_with_retrieved):
    def wrapped(settings={}):
        return parser_with_retrieved('vasprun', settings)

    return wrapped


@pytest.fixture
def parser_incomplete_output(parser_with_retrieved):
    """Fixture for tests with incomplete output"""
    default_settings = {
        'parser_settings': {
            'check_completeness': False,
            'critical_objects': [],
            'required_quantity': [],
        }
    }

    def wrapped(case, settings={}):
        if 'parser_settings' in settings:
            default_settings['parser_settings'].update(settings['parser_settings'])
        return parser_with_retrieved(case, default_settings)

    return wrapped


@pytest.fixture
def parser_with_outcar(parser_with_retrieved):
    def wrapped(settings={}):
        return parser_with_retrieved('outcar', settings)

    return wrapped


def test_parser_born(parser_with_retrieved):
    """Test parsing born effective charges"""
    parser, _ = parser_with_retrieved(
        'born_effective_charge',
        {
            'parser_settings': {
                'include_quantity': ['born_charges'],
                'check_completeness': False,
                'critical_objects': [],
            }
        },
    )
    assert 'born_charges' in parser.outputs


def test_parser_dielectrics(parser_incomplete_output):
    """Test parsing dielectrics"""
    parser, _ = parser_incomplete_output('dielectric')
    assert 'dielectrics' in parser.outputs


def test_parser_magnetization(parser_incomplete_output):
    """Test parsing dielectrics"""
    parser, _ = parser_incomplete_output('magnetization')
    assert 'misc' in parser.outputs
    assert parser.outputs['misc']['magnetization'][0] == 6.4424922


def test_parser_lepsilon(parser_incomplete_output):
    """Test parsing dielectrics"""
    parser, _ = parser_incomplete_output('lepsilon')
    assert 'misc' in parser.outputs
    assert 'epsilon' in parser.outputs['dielectrics'].get_arraynames()
    assert 'born_charges' in parser.outputs['born_charges'].get_arraynames()


def test_parser_localfield(parser_incomplete_output):
    """Test parsing dielectrics"""
    parser, _ = parser_incomplete_output('localfield')
    assert 'misc' in parser.outputs
    assert 'epsilon' in parser.outputs['dielectrics'].get_arraynames()
    assert 'born_charges' in parser.outputs['born_charges'].get_arraynames()


def test_parser_disp_details(parser_with_retrieved):
    """Test parsing elastic moduli and symmetries from OUTCAR"""
    parser, _ = parser_with_retrieved(
        'disp_details',
        {
            'parser_settings': {
                'include_quantity': ['symmetries', 'elastic_moduli'],
                'check_completeness': False,
                'critical_objects': [],
            }
        },
    )
    assert 'symmetries' in parser.outputs['misc']
    assert 'elastic_moduli' in parser.outputs['misc']
    data_dict = parser.outputs['misc'].get_dict()
    # Test
    test = np.array([1674.5786, 704.739, 704.739, -0.0, 0.0, 0.0])
    np.testing.assert_allclose(data_dict['elastic_moduli']['symmetrized'][0], test)
    test = np.array([0.0, 0.0, 0.0, -0.0, -0.0, 1122.6622])
    np.testing.assert_allclose(data_dict['elastic_moduli']['symmetrized'][5], test)
    test = np.array([705.0238, 1674.8491, 705.0238, -0.0, -0.0, 0.0])
    np.testing.assert_allclose(data_dict['elastic_moduli']['non_symmetrized'][1], test)
    test = np.array([-0.0078, -0.0495, 0.0147, 0.0, 1123.0829, -0.0])
    np.testing.assert_allclose(data_dict['elastic_moduli']['non_symmetrized'][4], test)
    test = np.array([704.739, 704.739, 1674.5786, -0.0, -0.0, 0.0])
    np.testing.assert_allclose(data_dict['elastic_moduli']['total'][2], test)
    test = np.array([-0.0, -0.0, -0.0, 775.8054, 0.0, -0.0])
    np.testing.assert_allclose(data_dict['elastic_moduli']['total'][3], test)

    assert data_dict['run_stats']
    assert data_dict['run_stats']['total_cpu_time_used'] == pytest.approx(89.795)
    assert data_dict['run_stats']['average_memory_used'] == pytest.approx(0.0)

    assert data_dict['run_status']['last_iteration_index'] == [15, 5]
    assert data_dict['run_status']['finished']
    assert data_dict['run_status']['ionic_converged']
    assert data_dict['run_status']['electronic_converged']
    assert data_dict['run_status']['nelm'] == 60
    assert data_dict['run_status']['nsw'] == 61
    assert data_dict['fermi_level'] == pytest.approx(6.17267267)
    assert np.amax(np.linalg.norm(data_dict['stress'], axis=1)) == pytest.approx(42.96872956444064)
    assert np.amax(np.abs(np.linalg.norm(data_dict['forces'], axis=1))) == pytest.approx(0.21326679)
    assert data_dict['total_energies']['energy_extrapolated'] == pytest.approx(-10.823296)


def test_vasprun_parsing(parser_with_vasprun):
    parser, _ = parser_with_vasprun(
        {'parser_settings': {'check_completeness': False, 'required_quantity': [], 'critical_objects': []}}
    )
    misc = parser.outputs['misc'].get_dict()
    shoud_exists = ['fermi_level', 'forces', 'stress', 'band_properties']
    for name in shoud_exists:
        assert name in misc


def test_outcar_parsing(parser_with_outcar):
    parser, _ = parser_with_outcar(
        {'parser_settings': {'check_completeness': False, 'required_quantity': [], 'critical_objects': []}}
    )
    misc = parser.outputs['misc'].get_dict()
    shoud_exists = ['run_stats', 'run_status']
    for name in shoud_exists:
        assert name in misc


def test_basic_run(parser_with_retrieved):
    parser, _ = parser_with_retrieved(
        'basic_run', {'parser_settings': {'include_quantity': ['born_charges'], 'critical_objects': []}}
    )

    misc = parser.outputs['misc'].get_dict()
    assert misc['band_properties']['cbm'] == pytest.approx(5.075)
    assert misc['band_properties']['vbm'] == pytest.approx(4.2811)
    assert misc['band_properties']['band_gap'] == pytest.approx(0.793899999999999)
    assert not misc['band_properties']['is_direct_gap']
    assert misc['version'] == '5.3.5'
    assert misc['total_energies']['energy_extrapolated'] == pytest.approx(-36.09616894)
    assert np.amax(np.linalg.norm(misc['stress'], axis=1)) == pytest.approx(8.50955439)
    assert np.amax(np.linalg.norm(misc['forces'], axis=1)) == pytest.approx(0.0)
    assert misc['run_status']['nelm'] == 60
    assert misc['run_status']['last_iteration_index'] == [1, 2]
    assert misc['run_status']['nsw'] == 0
    assert misc['run_status']['finished']
    assert misc['run_status']['ionic_converged'] is None
    assert misc['run_status']['electronic_converged']
    assert not misc['run_status']['consistent_nelm_breach']
    assert not misc['run_status']['contains_nelm_breach']
    assert misc['run_stats']['mem_usage_base'] == pytest.approx(30000.0)
    assert misc['run_stats']['mem_usage_nonl-proj'] == pytest.approx(7219.0)
    assert misc['run_stats']['mem_usage_fftplans'] == pytest.approx(776.0)
    assert misc['run_stats']['mem_usage_grid'] == pytest.approx(1605.0)
    assert misc['run_stats']['mem_usage_one-center'] == pytest.approx(124.0)
    assert misc['run_stats']['mem_usage_wavefun'] == pytest.approx(814.0)
    assert misc['run_stats']['maximum_memory_used'] == pytest.approx(95344.0)
    assert misc['run_stats']['average_memory_used'] == pytest.approx(0.0)
    assert misc['run_stats']['total_cpu_time_used'] == pytest.approx(20.463)
    assert misc['run_stats']['user_time'] == pytest.approx(11.266)
    assert misc['run_stats']['system_time'] == pytest.approx(9.197)
    assert misc['run_stats']['elapsed_time'] == pytest.approx(22.518)
    assert misc.get('magnetization') is None
    assert misc.get('site_magnetization') is None


def test_relax_run(parser_with_retrieved):
    parser, _ = parser_with_retrieved(
        'relax',
        {
            'parser_settings': {
                'include_quantity': ['born_charges'],
                'include_node': ['energies', 'trajectory'],
                'check_completeness': False,
                'required_quantity': [],
                'electronic_step_energies': True,
                'critical_objects': [],
            }
        },
    )

    array = parser.outputs['energies']
    energies_ext = array.get_array('energy_extrapolated')
    energies_ext_elec = array.get_array('energy_extrapolated_electronic')
    energies_elec_steps = array.get_array('electronic_steps')
    test_array_energies = [
        np.array(
            [
                163.37398579,
                14.26925896,
                -23.05190509,
                -34.91615104,
                -40.20080347,
                -42.18390876,
                -42.97469852,
                -43.31556073,
                -43.60169068,
                -43.61723125,
                -43.61871511,
                -43.61879751,
                -43.12548175,
                -42.90647187,
                -42.91031846,
                -42.91099027,
                -42.91111107,
                -42.91113348,
            ]
        ),
        np.array([-43.34236449, -43.31102002, -43.27768275, -43.27791002, -43.27761357, -43.27757545]),
        np.array([-43.40320524, -43.38084022, -43.36835045, -43.36666248, -43.36666583, -43.36649036, -43.36648855]),
        np.array([-43.37749056, -43.37749102, -43.37734414, -43.37734069]),
        np.array([-43.38117265, -43.38082881, -43.38063293, -43.38062479]),
        np.array([-43.38337336, -43.38334165]),
        np.array([-43.38778922, -43.38766017, -43.38752953, -43.38753003]),
        np.array([-43.38714489, -43.38708193]),
        np.array([-43.38640951, -43.38641449]),
        np.array([-43.3874799, -43.3871553, -43.38701949, -43.38701639]),
        np.array([-43.38790942, -43.38727062, -43.38700335, -43.38699488]),
        np.array([-43.38774394, -43.38773717]),
        np.array([-43.38984942, -43.3899134, -43.38988315]),
        np.array([-43.38988117, -43.3898822]),
        np.array([-43.39032165, -43.39017866, -43.39011239]),
        np.array([-43.39021044, -43.39020751]),
        np.array([-43.39034135, -43.39034244]),
        np.array([-43.39044466, -43.39044584]),
        np.array([-43.39084354, -43.39088709, -43.39087657]),
    ]
    test_array_steps = np.array([18, 6, 7, 4, 4, 2, 4, 2, 2, 4, 4, 2, 3, 2, 3, 2, 2, 2, 3])
    # Build a flattened array (not using flatten from NumPy as the content is staggered) and
    # test number of electronic steps per ionic step
    test_array_energies_flattened = np.array([])
    for ionic_step in test_array_energies:
        test_array_energies_flattened = np.append(test_array_energies_flattened, ionic_step)
    assert energies_ext_elec.shape == test_array_energies_flattened.shape
    np.testing.assert_allclose(test_array_energies_flattened, energies_ext_elec, atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(test_array_steps, energies_elec_steps, atol=0.0, rtol=1.0e-7)
    test_array_energies = np.array(
        [
            -0.00236637,
            -0.00048614,
            -0.00047201,
            -0.00043261,
            -0.00041668,
            -0.00042584,
            -0.00043637,
            -0.00042806,
            -0.00042762,
            -0.00043875,
            -0.00042731,
            -0.00042705,
            -0.00043064,
            -0.00043051,
            -0.00043161,
            -0.00043078,
            -0.00043053,
            -0.00043149,
            -0.00043417,
        ]
    )
    # Testing on VASP 5, where the extrapolated energy should be the following due to a bug
    with np.testing.assert_raises(AssertionError):
        np.testing.assert_allclose(test_array_energies, energies_ext, atol=0.0, rtol=1.0e-7)
    # Instead we correct and it should be
    test_array = np.array(
        [
            -42.911133,
            -43.277575,
            -43.366489,
            -43.377341,
            -43.380625,
            -43.383342,
            -43.38753,
            -43.387082,
            -43.386414,
            -43.387016,
            -43.386995,
            -43.387737,
            -43.389883,
            -43.389882,
            -43.390112,
            -43.390208,
            -43.390342,
            -43.390446,
            -43.390877,
        ]
    )
    np.testing.assert_allclose(test_array, energies_ext, atol=0.0, rtol=1.0e-7)

    traj = parser.outputs['trajectory']
    assert 'energy_extrapolated' in traj.get_arraynames()
    assert 'forces' in traj.get_arraynames()
    assert 'stress' in traj.get_arraynames()
    assert traj.get_array('forces').shape == (19, 8, 3)
    traj.get_stepids()
    traj.get_cells()
    traj.get_positions()
    traj.get_step_data(1)
    assert isinstance(traj.get_step_structure(1), orm.StructureData)


def test_basic(parser_with_retrieved):
    parser, _ = parser_with_retrieved(
        'basic',
        {
            'parser_settings': {
                'critical_objects': [],
                'check_completeness': False,
                'include_quantity': ['kpoints'],
                'required_quantity': [],
                'include_node': ['kpoints'],
                'check_errors': False,
            }
        },
    )
    assert 'kpoints' in parser.outputs
    kpoints = parser.outputs['kpoints']
    np.testing.assert_allclose(kpoints.get_kpoints()[0], np.array([0.0, 0.0, 0.0]), atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(
        kpoints.get_kpoints()[-1], np.array([0.42857143, -0.42857143, 0.42857143]), atol=0.0, rtol=1.0e-7
    )


def test_stream(parser_with_retrieved):
    """Test the functionality of the stream parser."""
    parser, _ = parser_with_retrieved(
        'stdout/out',
        {
            'parser_settings': {
                'critical_objects': [],
                'check_completeness': False,
                'include_quantity': ['stream'],
                'required_quantity': [],
            }
        },
    )
    misc_dict = parser.outputs['misc'].get_dict()
    assert misc_dict['notifications'][0]['name'] == 'ibzkpt'
    assert misc_dict['notifications'][0]['kind'] == 'ERROR'
    assert misc_dict['notifications'][0]['regex'] == 'internal error in subroutine IBZKPT'
    assert misc_dict['notifications'][1]['name'] == 'nostart'
    assert misc_dict['notifications'][1]['kind'] == 'ERROR'
    assert misc_dict['notifications'][1]['regex'] == 'vasp.'


def test_parser_exception(request, calc_with_retrieved):
    """
    This calculation has a missing eigenvalues section in the vasprun.xml
    """
    # This should work as the parser does not output the band by default
    # But the diagonsis information is missing so the erorr code is not zero
    settings_dict = {
        'parser_settings': {'check_completeness': False, 'critical_objects': ['vasprun.xml', 'vasp_output', 'OUTCAR']}
    }
    file_path = str(pathlib.Path(request.fspath).parent / '../test_data/basic_run_ill_format')
    node = calc_with_retrieved(file_path, settings_dict)
    _, output = VaspParser.parse_from_node(node, store_provenance=False, retrieved_temporary_folder=file_path)

    assert output.is_finished
    assert output.exit_status == 0

    # 1004 - node cannot be created as the quantity is missing
    settings_dict = {'parser_settings': {'check_completeness': True, 'include_node': ['bands']}}
    file_path = str(pathlib.Path(request.fspath).parent / '../test_data/basic_run_ill_format')
    node = calc_with_retrieved(file_path, settings_dict)
    _, output = VaspParser.parse_from_node(node, store_provenance=False, retrieved_temporary_folder=file_path)

    assert output.is_finished
    assert output.exit_status == 1004

    # 1002 - we explicitly require the eigenvalues to be present, but it is not
    settings_dict = {
        'parser_settings': {
            'critical_objects': [],
            'check_completeness': True,
            'required_quantity': ['eigenvalues'],
            'include_node': ['band'],
        }
    }
    file_path = str(pathlib.Path(request.fspath).parent / '../test_data/basic_run_ill_format')
    node = calc_with_retrieved(file_path, settings_dict)
    _, output = VaspParser.parse_from_node(node, store_provenance=False, retrieved_temporary_folder=file_path)

    assert output.is_finished
    assert output.exit_status == 1002


def test_notification_composer(parser_with_retrieved):
    """Test the NotificationComposer class"""

    parser, exit_code = parser_with_retrieved(
        'basic',
        {
            'parser_settings': {
                'critical_objects': [],
                'check_completeness': False,
                'include_quantity': ['kpoints'],
                'required_quantity': [],
            }
        },
        parse=False,
    )

    notifications = [{'name': 'edwav', 'kind': 'ERROR', 'message': 'Error in EDWAV'}]
    config = ParserSettingsConfig()
    composer = NotificationComposer(
        notifications,
        {},
        {
            'parameters': orm.Dict(dict={'nelect': 10}),
        },
        parser.exit_codes,
        critical_notifications=config.critical_notification_errors,
    )
    exit_code = composer.compose()
    assert exit_code.status == 703

    # BRMIX error but has NELECT defined in the input
    notifications = [{'name': 'brmix', 'kind': 'ERROR', 'message': 'Error in BRMIX'}]
    composer = NotificationComposer(
        notifications,
        {},
        {
            'parameters': orm.Dict(dict={'nelect': 10}),
        },
        parser.exit_codes,
        critical_notifications=config.critical_notification_errors,
    )
    exit_code = composer.compose()
    assert exit_code is None

    # BRMIX error but no NELECT tag
    composer = NotificationComposer(
        notifications,
        {},
        {'parameters': orm.Dict(dict={})},
        parser.exit_codes,
        critical_notifications=config.critical_notification_errors,
    )
    exit_code = composer.compose()
    assert exit_code.status == 703



"""
The ``INCAR`` parser interface.

Contains the parsing interfaces to parsevasp used to parse ``INCAR`` content.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from aiida import orm
from parsevasp.incar import Incar

from aiida_vasp.parsers.content_parsers.base import BaseFileParser

if TYPE_CHECKING:
    from io import TextIO


class IncarParser(BaseFileParser):
    """The parser interface that enables parsing of ``INCAR`` content.

    The parser is triggered by using the ``incar`` quantity key.

    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['incar']}

    PARSABLE_QUANTITIES = {
        'incar': {'inputs': [], 'name': 'incar', 'prerequisites': []},
    }

    def __init__(self, *args, validate_tags: bool = True, **kwargs) -> None:
        self._validate_tags = validate_tags
        super().__init__(*args, **kwargs)

    def _init_from_handler(self, handler: TextIO) -> None:
        """Initialize a ``parsevasp`` object of ``Incar`` using a file like handler.

        :param handler: A file like object that provides the necessary ``INCAR`` content to be parsed.
        :type handler: file-like object
        """

        try:
            self._content_parser = Incar(file_handler=handler, logger=self._logger, validate_tags=self._validate_tags)
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    def _init_from_data(self, data: orm.Dict) -> None:
        """Initialize using an AiiDA ``Dict`` instance.

        :param data: A valid AiiDA ``Dict`` object.
        :type data: object
        """

        if isinstance(data, orm.Dict):
            self._content_data = data
        else:
            raise TypeError('The supplied AiiDA data structure is not a Dict.')

    @property
    def incar(self) -> dict | None:
        """Return the parameters in the ``INCAR``.

        :returns: A dictionary containing the parameter tags as keys and its settings as values.
                  ``None`` is returned if the quantity can not be parsed.
        :rtype: dict or None
        """
        if self._content_parser is not None:
            params = self._content_parser.get_dict()
            return params
        return None

    def _content_data_to_content_parser(self) -> Incar:
        """Convert an AiiDA ``Dict`` to a content parser instance of ``Incar`` from ``parsevasp``.

        :returns: An instance of ``Incar`` from ``parsevasp``.
        :rtype: object
        """
        # Filter away None values from the dictionary - these are not valid for ``parsevasp``
        # This allow easier workflow control and parameters merging - setting a key to None means it should not be
        # set
        dictionary = {key: value for key, value in self._content_data.get_dict().items() if value is not None}

        # We brake hard if ``parsevasp`` fail here. If we can not write we will not try another parser.
        content_parser = Incar(incar_dict=dictionary, logger=self._logger, validate_tags=self._validate_tags)

        return content_parser



"""
Tools for using the Pymatgen library with aiida-vasp.
"""

from __future__ import annotations

import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional

try:
    import pymatgen.io.vasp as pvasp
except ImportError:
    raise ImportError('You need to install pymatgen to use this module.')


import gzip

from aiida import orm

from .aiida_utils import ensure_node_first_arg
from .export import export_vasp


@contextmanager
def temporary_folder() -> Generator[Path, None, None]:
    """Get a temporary folder and delete it after use."""
    tmpf = Path(tempfile.mkdtemp())
    yield tmpf
    shutil.rmtree(tmpf)


class PymatgenAdapator:
    """
    Adaptor for getting pymatgen objects from a VASP calculation/workflow
    This work by first exporting the calculation to a temporary folder and then parsing the files using pymmatgen.

    Some of the pymatgen objects does not have the from_dict method implemented as required by MSONable.
    Hence, they can only be reconstructed as a dictionary.
    """

    FILES = {
        'potcar': ('Potcar', 'POTCAR'),
        'vasprun': ('Vasprun', 'vasprun.xml'),
        'kpoints': ('Kpoints', 'KPOINTS'),
        'ibzkpt': ('Kpoints', 'IBZKPT'),
        'incar': ('Incar', 'INCAR'),
        'outcar': ('Outcar', 'OUTCAR'),
        'contcar': ('Poscar', 'CONTCAR'),
        'poscar': ('Poscar', 'POSCAR'),
        'chgcar': ('Chgcar', 'CHGCAR'),
    }
    # Classes where from_dict is not implemented but still MSONable
    NO_RECONSTRUCT = ['vasprun', 'outcar', 'chgcar']

    def __init__(self, node: orm.CalcJobNode, store_cache: bool = True) -> None:
        """Adaptor for getting pymatgen objects from a VASP calculation/workflow"""
        self.node = node
        self.pmg_objects = {}
        self.cache = {}
        self.store_cache = store_cache

    def _parse_full(self, file_names: Optional[List[str]] = None) -> None:
        """
        Parse all files and save to the pmg_objects attribute
        The assumption is that exporting the calculation folder is the slowest part of the process.
        """
        if file_names is None:
            file_names = self.FILES
        else:
            file_names = {key: self.FILES[key] for key in file_names}

        with temporary_folder() as tmpf:
            export_vasp(self.node, tmpf)
            for name, (cls_name, file) in file_names.items():
                # Instantiate the pymatgen object
                cls = getattr(pvasp, cls_name)
                if not Path(tmpf / file).is_file():
                    continue
                # Try use the from_file method if it exists
                fname = str(tmpf / file)
                if hasattr(cls, 'from_file'):
                    try:
                        obj = cls.from_file(fname)
                    # Skip if the file is not found or the parsing fails
                    except Exception:
                        continue
                # If using Vasprun, try to parse the potcar file but fall back when needed
                elif cls == pvasp.Vasprun:
                    try:
                        obj = cls(fname)
                    except ValueError:
                        obj = cls(fname, parse_potcar=False)
                else:
                    obj = cls(fname)
                self.pmg_objects[name] = obj

    def export_files(self, dst: str | Path) -> None:
        """Export the VASP calculation files to a destination folder"""
        export_vasp(self.node, dst)

    def _get_pmg_object(self, name: str) -> Any:
        """
        Get a pymatgen object

        1. If we can find the object in parsed object , then just return it.
        2. If it is not already parsed, try to load the cache (stored in the extras)
        3. Otherwise, try to export and parse from the files explicitly. (slow)

        :param name: Name of the object to get (e.g. 'vasprun', 'outcar', 'poscar', 'incar', 'kpoints', 'ibzkpt')
        """
        # We already parsed the calculation, so we can just return the object
        # Since we have access it - save it to the cache
        if name in self.pmg_objects:
            if name not in self.cache:
                self.cache[name + '_dict'] = self.pmg_objects[name].as_dict()
            return self.pmg_objects[name]

        # Check if the object was accessed before and previously save to the cache
        if not self.cache:
            self.cache = self.node.base.extras.get('pmg_cache', {})

        if name + '_dict' in self.cache and name not in self.NO_RECONSTRUCT:
            # Already in the cache - return the object
            return getattr(pvasp, self.FILES[name][0]).from_dict(self.cache[name])
        else:
            # Not in the cache
            if name not in self.pmg_objects:
                # We have not paresed the calculation yet - do it now
                self._parse_full()
            # Get the object
            obj = self.pmg_objects[name]
            # Save the parsed object to the cache
            self.cache[name + '_dict'] = obj.as_dict()
            return obj

    def _get_pmg_dict(self, name: str) -> Dict[str, Any]:
        """
        Get a pymatgen object as a dictionary
        """
        if not self.cache:
            self.cache = self.node.base.extras.get('pmg_cache', {})
        if name + '_dict' in self.cache:
            return self.cache[name + '_dict']
        return self._get_pmg_object(name).as_dict()

    def _flush_cache(self) -> None:
        """Close the adaptor and save the cache"""
        self.node.base.extras.set('pmg_cache', self.cache)

    def __enter__(self) -> 'PymatgenAdapator':
        """Enter the adaptor"""
        return self

    def __exit__(self, *args, **kwargs) -> None:
        if self.store_cache:
            self._flush_cache()

    @property
    def vasprun(self) -> pvasp.Vasprun:
        """Return the Vasprun object"""
        return self._get_pmg_object('vasprun')

    @property
    def vasprun_dict(self) -> Dict[str, Any]:
        """Return the Vasprun object as dictionary (will trigger caching)"""
        return self._get_pmg_dict('vasprun')

    @property
    def outcar(self) -> pvasp.Outcar:
        """Return the Outcar object"""
        return self._get_pmg_object('outcar')

    @property
    def outcar_dict(self) -> Dict[str, Any]:
        """Return the Outcar object as dictionary (will trigger caching)"""
        return self._get_pmg_dict('outcar')

    @property
    def poscar(self) -> pvasp.Poscar:
        """Return the Poscar object"""
        return self._get_pmg_object('poscar')

    @property
    def poscar_dict(self) -> Dict[str, Any]:
        """Return the Poscar object as dictionary (will trigger caching)"""
        return self._get_pmg_dict('poscar')

    @property
    def incar(self) -> pvasp.Incar:
        """Return the Incar object"""
        return self._get_pmg_object('incar')

    @property
    def incar_dict(self) -> Dict[str, Any]:
        """Return the Incar object as dictionary"""
        return self._get_pmg_dict('incar')

    @property
    def kpoints(self) -> pvasp.Kpoints:
        """Return the Kpoints object"""
        return self._get_pmg_object('kpoints')

    @property
    def kpoints_dict(self) -> Dict[str, Any]:
        """Return the Kpoints object as dictionary"""
        return self._get_pmg_dict('kpoints')

    @property
    def ibzkpt(self) -> pvasp.Kpoints:
        """Return the IBZKPT object"""
        return self._get_pmg_object('ibzkpt')

    @property
    def ibzkpt_dict(self) -> Dict[str, Any]:
        """Return the IBZKPT object as dictionary"""
        return self._get_pmg_dict('ibzkpt')

    def save_msonable(self, name: str, obj: Any) -> None:
        """Save msonable object to the node extras"""
        dobj = obj.as_dict()
        assert '@module' in dobj
        self.node.base.extras.set(f'pmg_cache_{name}', dobj)

    def load_msonable(self, name: str) -> Any:
        """Load msonable object from the node extras"""
        from monty.json import MontyDecoder  # noqa: PLC0415

        return MontyDecoder().process_decoded(self.node.base.extras.get(f'pmg_cache_{name}'))


@ensure_node_first_arg
def get_vasprun(node: Any, store_cache: bool = True) -> pvasp.Vasprun:
    """Return the Vasprun object"""
    return PymatgenAdapator(node, store_cache=store_cache).vasprun


@ensure_node_first_arg
def get_outcar(node: Any, store_cache: bool = True) -> pvasp.Outcar:
    """Return the OUTCAR object"""
    return PymatgenAdapator(node, store_cache=store_cache).outcar


@ensure_node_first_arg
def get_incar(node: Any, store_cache: bool = True) -> pvasp.Incar:
    """Return the INCAR object"""
    return PymatgenAdapator(node, store_cache=store_cache).incar


@ensure_node_first_arg
def get_kpoints(node: Any, store_cache: bool = True) -> pvasp.Kpoints:
    """Return the Kpoints object"""
    return PymatgenAdapator(node, store_cache=store_cache).kpoints


@ensure_node_first_arg
def get_ibzkpt(node: Any, store_cache: bool = True) -> pvasp.Kpoints:
    """Return the Kpoints object using the IBZKPT file"""
    return PymatgenAdapator(node, store_cache=store_cache).ibzkpt


def convert_pymatgen_potcar_folder(src: Path | str, dst: Path | str) -> None:
    """
    Convert pymatgen potcar folder to a structure used by aiida-vasp

    :param src: Path to the pymatgen potcar folder
    :param dst: Path to the aiida-vasp potcar folder

    :return: None
    """

    src = Path(src)
    dst = Path(dst)
    for fpath in Path(src).glob('POTCAR.*.gz'):
        symbol = fpath.name.split('.')[1]
        folder = dst / symbol
        folder.mkdir(exist_ok=True, parents=True)
        # unzip the file
        with gzip.open(fpath, 'rb') as f_in:
            with (folder / 'POTCAR').open('wb') as f_out:
                shutil.copyfileobj(f_in, f_out)



"""Unit test the POTCAR AiiDA data structures."""

import pathlib
import subprocess as sp
import tarfile

import pytest
from aiida import orm
from aiida.common.exceptions import NotExistent, UniquenessError

from aiida_vasp.data.potcar import PotcarData, PotcarFileData, PotcarGroup, migrate_potcar_group


def test_creation(aiida_profile_clean, potcar_node_pair):
    """Test creating a data node pair."""
    potcar_node = PotcarData.find_one(symbol='As')
    file_node = potcar_node.find_file_node()
    assert potcar_node.pk == potcar_node_pair['potcar'].pk
    assert file_node.pk == potcar_node_pair['file'].pk


def test_hashing(aiida_profile_clean, data_path, read_file):
    """Ensure the file and content sha512 hash equivalently for the same POTCAR."""
    potcar_file_cls = PotcarFileData
    potcar_path = ['potcar', 'As', 'POTCAR']

    file_sha512 = potcar_file_cls.get_file_sha512(data_path(*potcar_path))
    content_sha512 = potcar_file_cls.get_contents_sha512(read_file(*potcar_path, mode='rb'))
    assert file_sha512 == content_sha512


# pylint: disable=protected-access
def test_store_duplicate(aiida_profile_clean, potcar_node_pair, data_path):
    """
    Storing a duplicate POTCAR node must fail.

    Uniqueness constraints to test for:

        * ``sha512`` attribute must be unique
        * the combination of all other attributes must be unique
    """
    potcar_path = data_path('potcar', 'As', 'POTCAR')

    file_node = PotcarFileData(file=potcar_path)
    file_node.base.attributes.set('sha512', 'foo')
    with pytest.raises(UniquenessError):
        file_node.store()

    file_node = PotcarFileData(file=potcar_path)
    file_node.base.attributes.set('symbol', 'Ta')
    with pytest.raises(UniquenessError):
        file_node.store()

    data_node = PotcarData(potcar_file_node=potcar_node_pair['file'])
    data_node.base.attributes.set('sha512', 'foo')
    with pytest.raises(UniquenessError):
        data_node.store()

    data_node = PotcarData(potcar_file_node=potcar_node_pair['file'])
    data_node.base.attributes.set('symbol', 'Ta')
    with pytest.raises(UniquenessError):
        data_node.store()

    assert PotcarData.find_one(symbol='As')
    assert PotcarFileData.find_one(symbol='As')


def test_export_import(aiida_profile_clean, potcar_node_pair, tmp_path, data_path):
    """Exporting and importing back may not store duplicates."""
    tempfile = tmp_path / 'potcar.aiida'

    sp.call(['verdi', 'export', 'create',
             '-n', str(potcar_node_pair['file'].pk),
             '-n', str(potcar_node_pair['potcar'].pk), str(tempfile)])  # yapf: disable

    # import with same uuid
    sp.call(['verdi', 'import', str(tempfile)])
    assert PotcarData.find_one(symbol='As')
    assert PotcarFileData.find_one(symbol='As')

    # import with different uuid
    sp.call(['verdi', 'import', data_path('potcar', 'export.aiida')])
    assert PotcarData.find_one(symbol='As')
    assert PotcarFileData.find_one(symbol='As')


def test_exists(aiida_profile_clean, potcar_node_pair):
    assert PotcarFileData.exists(element='As')
    assert not PotcarData.exists(element='Xe')


def test_find(aiida_profile_clean, potcar_node_pair):
    assert PotcarData.find_one(element='As').uuid == potcar_node_pair['potcar'].uuid
    with pytest.raises(NotExistent):
        _ = PotcarFileData.find_one(element='Xe')


def test_file_get_content(aiida_profile_clean, potcar_node_pair, data_path):
    file_node_as = potcar_node_pair['file']
    original_file = pathlib.Path(data_path(file_node_as.original_file_name))
    assert original_file.read_text(encoding='utf8') == file_node_as.get_content().decode()


def test_file_get_or_create(aiida_profile_clean, potcar_node_pair, data_path):
    """Test get_or_create of PotcarFileData."""
    potcar_as_path = data_path('potcar', 'As', 'POTCAR')
    potcar_file_cls = PotcarFileData
    file_as = potcar_node_pair['file']
    node_file_as, created_file_as = potcar_file_cls.get_or_create(potcar_as_path)
    assert not created_file_as
    assert file_as.pk == node_file_as.pk

    potcar_in_path = data_path('potcar', 'In_d', 'POTCAR')
    node_file_in, created_file_in = potcar_file_cls.get_or_create(potcar_in_path)
    assert created_file_in
    assert potcar_file_cls.exists(sha512=node_file_in.sha512)


def test_potcar_get_or_create(aiida_profile_clean, potcar_node_pair, data_path):
    """Test get_or_create method of PotcarData."""
    potcar_cls = PotcarData
    file_cls = PotcarFileData
    file_as = potcar_node_pair['file']
    potcar_as = potcar_node_pair['potcar']
    node_potcar_as, created_potcar_as = potcar_cls.get_or_create(file_as)
    assert not created_potcar_as
    assert potcar_as.pk == node_potcar_as.pk

    potcar_in_path = data_path('potcar', 'In_d', 'POTCAR')
    node_potcar_in, created_potcar_in = potcar_cls.get_or_create(file_cls(file=potcar_in_path))
    assert created_potcar_in
    assert potcar_cls.exists(sha512=node_potcar_in.sha512)


def test_potcar_from_file(aiida_profile_clean, data_path):
    """Test creating a node pair from a file, creating the data node first."""
    potcar_cls = PotcarData
    _, created = potcar_cls.get_or_create_from_file(data_path('potcar', 'As', 'POTCAR'))
    assert created
    _, created = potcar_cls.get_or_create_from_file(data_path('potcar', 'As', 'POTCAR'))
    assert not created


def test_potcar_from_structure(aiida_profile_clean, upload_potcar, potcar_family_name):
    """Test getting POTCARS from a family for a structure."""
    indium_2 = orm.StructureData()
    indium_2.append_atom(position=[0, 0, 0], symbols='In')
    indium_2.append_atom(position=[1, 0, 0], symbols='In', name='In1')
    in2_potcars = PotcarData.get_potcars_from_structure(
        indium_2, potcar_family_name, mapping={'In': 'In_d', 'In1': 'In_d'}
    )
    assert set(in2_potcars.keys()) == {'In', 'In1'}
    in_d_potcar = PotcarData.find(family_name=potcar_family_name, full_name='In_d')[0]
    assert in2_potcars['In'].uuid == in_d_potcar.uuid == in2_potcars['In1'].uuid


def test_upload(aiida_profile_clean, temp_pot_folder, data_path):
    """Test uploading a family of POTCAR files."""
    family_name = 'test_family'
    family_desc = 'Test Family'
    potcar_cls = PotcarData
    pot_dir = str(temp_pot_folder)
    potcar_ga = pathlib.Path(data_path('potcar')) / 'Ga'
    assert not potcar_ga.exists()

    potcar_cls.upload_potcar_family(pot_dir, family_name, family_desc)

    assert potcar_cls.exists(element='In')
    assert potcar_cls.exists(element='As')
    assert potcar_cls.exists(element='Ga')
    assert not potcar_ga.exists()

    # this is supposed to return only one group, however it returns 8 (= number of uploaded files)
    assert [g.label for g in potcar_cls.get_potcar_groups()] == [family_name]

    assert len(potcar_cls.get_potcar_group(family_name).nodes) >= 3

    with pytest.raises(ValueError):
        potcar_cls.upload_potcar_family(pot_dir, family_name, stop_if_existing=True)
    assert not potcar_ga.exists()

    num_files, num_added, num_uploaded = potcar_cls.upload_potcar_family(
        pot_dir, family_name + '_new', family_desc, stop_if_existing=False
    )
    assert num_files >= 3
    assert num_added >= 3
    assert num_uploaded == 0
    assert not potcar_ga.exists()


def test_export_family_folder(aiida_profile_clean, upload_potcar, potcar_family_name, tmp_path):
    """Test exporting to folder."""
    export_dir = tmp_path / 'export'
    export_dir.mkdir()
    potcar_cls = PotcarData

    # Check that dry run works and does not leave anything in the directory
    potcar_cls.export_family_folder(potcar_family_name, path=export_dir, dry_run=True)
    elements = []
    for item in export_dir.iterdir():
        elements.append(item)
    assert not elements

    # Start check for actual export
    files = potcar_cls.export_family_folder(potcar_family_name, path=export_dir, dry_run=False)
    family_dir = export_dir / potcar_family_name
    subdirs = set(str(subpath.name) for subpath in family_dir.iterdir())
    assert set(['Ga', 'As', 'In_d']).issubset(subdirs)
    potcar_ga = family_dir / 'Ga' / 'POTCAR'
    assert potcar_ga in files
    assert len(files) >= 3
    assert potcar_ga.is_file()
    assert 'TITEL' in potcar_ga.read_text()

    new_dir = export_dir / 'new_dir'
    potcar_cls.export_family_folder(potcar_family_name, path=new_dir, dry_run=False)
    assert new_dir.exists()


def test_export_family_archive(aiida_profile_clean, upload_potcar, potcar_family_name, tmp_path):
    """Test exporting to archive."""
    export_dir = tmp_path / 'export'
    export_dir.mkdir()
    potcar_cls = PotcarData

    # Check that dry run works and does not leave anything in the directory
    potcar_cls.export_family_archive(potcar_family_name, path=export_dir, dry_run=True)
    elements = []
    for item in export_dir.iterdir():
        elements.append(item)
    assert not elements

    # Start check for actual export
    ar_path, _ = potcar_cls.export_family_archive(potcar_family_name, path=export_dir, dry_run=False)
    with tarfile.open(str(ar_path)) as archive:
        assert set(['Ga/POTCAR', 'As/POTCAR', 'In_d/POTCAR']).issubset(set(archive.getnames()))
        potcar_in = archive.extractfile('In_d/POTCAR')
        content = potcar_in.read()
        assert b'TITEL' in content


def test_create_equivalence(upload_potcar, read_file):
    """Create from file (during upload) and from contents and ensure equivalence."""
    potcar_file_cls = PotcarFileData
    potcar_path = ['potcar', 'As', 'POTCAR']
    potcar_file, created = potcar_file_cls.get_or_create_from_contents(read_file(*potcar_path, mode='rb'))
    assert not created
    assert potcar_file.sha512 == potcar_file_cls.find_one(element='As').sha512
    assert potcar_file.uuid == potcar_file_cls.find_one(element='As').uuid

    potcar_cls = PotcarData
    potcar, created = potcar_cls.get_or_create_from_contents(read_file(*potcar_path, mode='rb'))
    assert not created
    assert potcar.sha512 == potcar_cls.find_one(element='As').sha512
    assert potcar.uuid == potcar_cls.find_one(element='As').uuid


def test_get_poctcars_dict(upload_potcar, potcar_family_name, potcar_mapping):
    """Test the keys are the same as the input element names."""
    potcar_cls = PotcarData
    elements = potcar_mapping.keys()
    potcar_dict = potcar_cls.get_potcars_dict(elements=elements, family_name=potcar_family_name, mapping=potcar_mapping)
    assert set(potcar_dict.keys()) == set(elements)
    assert [potcar_dict[element].full_name for element in elements] == [potcar_mapping[element] for element in elements]


def test_family_migrate(upload_potcar, legacy_potcar_family):
    """Test the migration from OLD potcar family to the new ones"""

    old_family_name, legacy_group_class = legacy_potcar_family
    legacy_group = legacy_group_class.collection.get(label=old_family_name)
    migrate_potcar_group()

    # Old group should still be there
    assert legacy_group_class.collection.get(label=old_family_name)

    migrated = PotcarGroup.collection.get(label=old_family_name)
    uuids_original = {node.uuid for node in legacy_group.nodes}
    uuids_migrated = {node.uuid for node in migrated.nodes}
    assert uuids_migrated == uuids_original


def test_old_style_detect(aiida_profile_clean, potcar_family_name, potcar_mapping, legacy_potcar_family):
    """Test the assestion that the potcars are found old in the legacy group not the new"""
    potcar_cls = PotcarData
    elements = potcar_mapping.keys()
    new_group = PotcarGroup.collection.get(label=potcar_family_name)
    new_group.label += '_'

    # Change the name of the legacy group to the new one so it will be matched
    legacy_group_label, legacy_group_class = legacy_potcar_family
    legacy_group = legacy_group_class.collection.get(label=legacy_group_label)
    legacy_group.label = potcar_family_name

    # The raise Value Error should contain hints for migrate
    with pytest.raises(NotExistent, match=r'.*found in a legacy group.*'):
        potcar_dict = potcar_cls.get_potcars_dict(
            elements=elements,
            family_name=potcar_family_name,
            mapping=potcar_mapping,
            auto_migrate=False,
        )

    # Change the name back and the test should now pass
    new_group.label = new_group.label[:-1]
    potcar_dict = potcar_cls.get_potcars_dict(
        elements=elements,
        family_name=potcar_family_name,
        mapping=potcar_mapping,
        auto_migrate=False,
    )
    assert set(potcar_dict.keys()) == set(elements)
    assert [potcar_dict[element].full_name for element in elements] == [potcar_mapping[element] for element in elements]

    # Test the auto-migration logic
    # Change the name again, so the only the old group matches
    new_group.label += '_'
    with pytest.raises(NotExistent):
        PotcarGroup.collection.get(label=potcar_family_name)

    # but as long as we do auto migrate it would be fine
    potcar_dict = potcar_cls.get_potcars_dict(
        elements=elements, family_name=potcar_family_name, mapping=potcar_mapping, auto_migrate=True
    )
    assert set(potcar_dict.keys()) == set(elements)
    assert [potcar_dict[element].full_name for element in elements] == [potcar_mapping[element] for element in elements]
    # Validate the migrate group
    migrated = PotcarGroup.collection.get(label=potcar_family_name)
    uuids_original = {node.uuid for node in legacy_group.nodes}
    uuids_migrated = {node.uuid for node in migrated.nodes}
    assert uuids_migrated == uuids_original



"""
The ``EIGENVAL`` parser interface.

Contains the parsing interfaces to parsevasp used to parse ``EIGENVAL`` content.
"""

# pylint: disable=abstract-method
from typing import TextIO

from parsevasp.eigenval import Eigenval

from aiida_vasp.parsers.content_parsers.base import BaseFileParser


class EigenvalParser(BaseFileParser):
    """The parser interface that enables parsing of ``EIGENVAL`` content.

    The parser is triggered by using the ``eigenval-eigenvalues`` and/or ``eigenval-kpoints`` quantity key.
    The quantity keys ``eigenvalues`` and ``kpoints`` will on the other hand parse the
    eigenvalues and/or kpoints using the XML parser.

    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['eigenval-eigenvalues', 'eigenval-kpoints']}

    PARSABLE_QUANTITIES = {
        'eigenval-eigenvalues': {
            'inputs': [],
            'name': 'eigenvalues',
            'prerequisites': [],
        },
        'eigenval-kpoints': {
            'inputs': ['structure'],
            'name': 'kpoints',
            'prerequisites': ['structure'],
        },
    }

    def _init_from_handler(self, handler: TextIO) -> None:
        """Initialize a ``parsevasp`` object of ``Eigenval`` using a file like handler.

        :param handler: A file like object that provides the necessary ``EIGENVAL`` content to be parsed.
        :type handler: file-like object
        """

        try:
            self._content_parser = Eigenval(file_handler=handler, logger=self._logger)
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    @property
    def eigenvalues(self):
        """Return the eigenvalue and metadata in the ``EIGENVAL``.

        :returns: A dict containing the keys ``eigenvalues`` and ``metadata``, which contain
                  a NumPy array of the eigenvalues and a dict with metadata, respectively.
        :rtype: dict
        """

        metadata = self._content_parser.get_metadata()
        _eigenvalues = self._content_parser.get_eigenvalues()
        eigenvalues = {'eigenvalues': _eigenvalues, 'metadata': metadata}

        return eigenvalues

    @property
    def kpoints(self):
        """Return the kpoints and metadata in the ``EIGENVAL``.

        :returns: A dict containing the keys ``kpoints`` and ``metadata``, which contain
                  a NumPy array of the k-points and a dict with metadata, respectively.
        :rtype: dict
        """

        metadata = self._content_parser.get_metadata()
        _kpoints = self._content_parser.get_kpoints()
        kpoints = {'kpoints': _kpoints, 'metadata': metadata}

        return kpoints



"""
Call script to calculate the total energies for one volume of standard silicon.

This particular call script set up a standard calculation that execute a calculation for
the fcc silicon structure.
"""
from aiida import load_profile

# pylint: disable=too-many-arguments, invalid-name, import-outside-toplevel
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import submit
from aiida.orm import Bool, Code, Str
from aiida.plugins import DataFactory, WorkflowFactory

load_profile()


def get_structure(label):
    from aiida.orm import QueryBuilder

    qb = QueryBuilder()
    qb.append(DataFactory('structure'), filters={'label': {'==': label}}, tag='structure')
    # Pick any structure with this label, here, just the first
    return qb.all()[0][0]


def main(code_string, incar, kmesh, structure, potential_family, potential_mapping, options):
    """Main method to setup the calculation."""

    # First, we need to fetch the AiiDA datatypes which will
    # house the inputs to our calculation
    dict_data = DataFactory('dict')
    kpoints_data = DataFactory('array.kpoints')

    # Then, we set the workchain you would like to call
    workchain = WorkflowFactory('vasp.vasp')

    # And finally, we declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()

    # Organize settings
    settings.parser_settings = {'add_dos': True}

    # Set inputs for the following WorkChain execution
    # Code
    inputs.code = Code.get_from_string(code_string)
    # Structure
    inputs.structure = structure
    # k-points grid density
    kpoints = kpoints_data()
    kpoints.set_kpoints_mesh(kmesh)
    inputs.kpoints = kpoints
    # Parameters
    inputs.parameters = dict_data(dict=incar)
    # Potential family and their mapping between element and potential type to use
    inputs.potential_family = Str(potential_family)
    inputs.potential_mapping = dict_data(dict=potential_mapping)
    # Options
    inputs.options = dict_data(dict=options)
    # Settings
    inputs.settings = dict_data(dict=settings)
    # Workchain related inputs, in this case, give more explicit output to report
    inputs.verbose = Bool(True)
    # Submit the workchain with the set inputs
    submit(workchain, **inputs)


if __name__ == '__main__':
    # Code_string is chosen among the list given by 'verdi code list'
    CODE_STRING = 'vasp@mycluster'

    # INCAR equivalent
    # Set input parameters
    INCAR = {'incar': {'encut': 240, 'ismear': -5, 'lorbit': 11}}

    # KPOINTS equivalent
    # Set kpoint mesh
    KMESH = [21, 21, 21]

    # POTCAR equivalent
    # Potential_family is chosen among the list given by
    # 'verdi data vasp.potcar listfamilies'
    POTENTIAL_FAMILY = 'PBE.54'
    # The potential mapping selects which potential to use, here we use the standard
    # for silicon, this could for instance be {'Si': 'Si_GW'} to use the GW ready
    # potential instead
    POTENTIAL_MAPPING = {'Si': 'Si'}

    # jobfile equivalent
    # In options, we typically set scheduler options.
    # See https://aiida.readthedocs.io/projects/aiida-core/en/latest/scheduler/index.html
    # AttributeDict is just a special dictionary with the extra benefit that
    # you can set and get the key contents with mydict.mykey, instead of mydict['mykey']
    OPTIONS = AttributeDict()
    OPTIONS.account = ''
    OPTIONS.qos = ''
    OPTIONS.resources = {'num_machines': 1, 'num_mpiprocs_per_machine': 8}
    OPTIONS.queue_name = ''
    OPTIONS.max_wallclock_seconds = 3600
    OPTIONS.max_memory_kb = 2000000

    # POSCAR equivalent
    # Set the silicon structure
    STRUCTURE = get_structure('silicon_at_3_9')

    main(CODE_STRING, INCAR, KMESH, STRUCTURE, POTENTIAL_FAMILY, POTENTIAL_MAPPING, OPTIONS)



from .common.builder_updater import *

__all__ = (
    'VaspBandUpdater',
    'VaspBuilderUpdater',
    'VaspConvUpdater',
    'VaspHybridBandUpdater',
    'VaspNEBUpdater',
    'VaspPresetConfig',
    'VaspRelaxUpdater',
)



"""
POTCAR parser.

The file parser that handles the parsing of POTCAR files. Also contains methods to
find, import, compose and write POTCAR files.
"""

from __future__ import annotations

import os
import re
from itertools import groupby
from pathlib import Path
from typing import Any, TextIO

from aiida import orm
from parsevasp.potcar import Potcar

from aiida_vasp.data.potcar import PotcarData, PotcarFileData
from aiida_vasp.parsers.content_parsers.base import BaseFileParser
from aiida_vasp.utils.delegates import delegate_method_kwargs


class PotcarParser(BaseFileParser):
    """A lightweight interface that provides access to POTCAR metadata parsing.

    Similar to the other content parser for VASP in structure, but only used directly in the POTCAR
    handling logic.

    """

    DEFAULT_SETTINGS = {}

    PARSABLE_QUANTITIES = {}

    def _init_from_handler(self, handler: TextIO) -> None:
        """Initialize using a file like handler.

        :param handler: A file like object that provides the necessary content to be parsed.
        :type handler: file-like object
        """

        try:
            self._content_parser = Potcar(file_handler=handler, logger=self._logger)
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    @property
    def metadata(self) -> Potcar:
        """Return the metadata Potcar instance."""
        return self._content_parser

    def _init_from_data(self, data: dict) -> None:
        """No need to init from an AiiDA data structure."""
        raise NotImplementedError('PotcarParser does not implement a _init_from_data() method.')

    def _content_data_to_content_parser(self) -> None:
        """Since no need to accept AiiDA data structure, no need to convert it."""
        raise NotImplementedError('PotcarParser does not implement a _content_data_to_content_parser() method.')


class PotcarIo:
    """
    Deals with VASP input output of POTCAR files.

    Instantiate with one of the following kwargs:

    :param path: (string) absolute path to the POTCAR file
    :param potcar_node: a PotcarData node
    :param potcar_file_node: a PotcarFileNode
    :param contents: a string with the POTCAR content
    """

    def __init__(self, **kwargs: Any) -> None:
        """Init from Potcar object or delegate to kwargs initializers."""
        self.potcar_obj = None
        self.sha512 = None
        self.init_with_kwargs(**kwargs)

    @delegate_method_kwargs(prefix='_init_with_')
    def init_with_kwargs(self, **kwargs: Any) -> None:
        """Delegate initialization to _init_with - methods."""

    def _init_with_path(self, file_path: str | Path) -> None:
        """Initialize with a path."""
        node, _ = PotcarData.get_or_create_from_file(file_path=file_path)
        self.sha512 = node.sha512

    def _init_with_potcar_file_node(self, node: PotcarFileData) -> None:
        """Initialize with an existing potential file node."""
        self.sha512 = node.sha512

    def _init_with_potcar_node(self, node: PotcarData) -> None:
        """Initialize with an existing potential node."""
        self._init_with_potcar_file_node(node.find_file_node())

    def _init_with_contents(self, contents: str) -> None:
        """Initialize with a string."""
        from aiida_vasp.data.potcar import PotcarData  # noqa: PLC0415

        try:
            contents = contents.encode('utf-8')
        except AttributeError:
            pass
        node, _ = PotcarData.get_or_create_from_contents(contents)
        self.sha512 = node.sha512

    @property
    def file_node(self) -> PotcarFileData:
        from aiida_vasp.data.potcar import PotcarData  # noqa: PLC0415

        return PotcarData.find_one(sha512=self.sha512).find_file_node()

    @property
    def node(self) -> PotcarData:
        from aiida_vasp.data.potcar import PotcarData  # noqa: PLC0415

        return PotcarData.find_one(sha512=self.sha512)

    @property
    def content(self) -> bytes:
        return self.file_node.get_content()

    @classmethod
    def from_(cls, potcar: str | Path | PotcarData | PotcarFileData | PotcarIo) -> PotcarIo:
        """Determine the best guess at how the input represents a POTCAR file and construct
        a PotcarIo instance based on that."""
        from aiida_vasp.data.potcar import PotcarData, PotcarFileData  # noqa: PLC0415

        if isinstance(potcar, (str, os.PathLike)):
            try:
                path_exists = Path(potcar).exists()
            except OSError:
                # We failed possibly due to a too long filename or that the potcar content is in fact
                # potcar content, revert to the os module to check if it exists
                path_exists = os.path.exists(potcar)
            if path_exists:
                potcar = cls(path=potcar)
            else:
                potcar = cls(contents=potcar)
        elif isinstance(potcar, PotcarData):
            potcar = cls(potcar_node=potcar)
        elif isinstance(potcar, PotcarFileData):
            potcar = cls(potcar_file_node=potcar)
        elif isinstance(potcar, PotcarIo):
            pass
        else:
            raise TypeError(f'Invalid type of potcar: {type(potcar)}')
        return potcar

    def __eq__(self, other: PotcarIo) -> bool:
        return self.sha512 == other.sha512

    def __hash__(self) -> int:
        return hash(self.sha512)


class MultiPotcarIo:
    """
    Handle file i/o for POTCAR files with one or more potentials.
    """

    def __init__(self, potcars: list[Any] | None = None) -> None:
        self._potcars = []
        if potcars:
            for potcar in potcars:
                self.append(PotcarIo.from_(potcar))

    def append(self, potcar: Any) -> None:
        self._potcars.append(PotcarIo.from_(potcar))

    def write(self, path: str | Path) -> None:
        path = Path(path)
        with path.open('wb') as dest_fo:
            for potcar in self._potcars:
                dest_fo.write(potcar.content)

    @classmethod
    def read(cls, path: str | Path) -> MultiPotcarIo:
        """Read a POTCAR file that may contain one or more potentials into a list of PotcarIo objects."""
        potcars = cls()
        path = Path(path)
        with path.open('r', encoding='utf8') as potcar_fo:
            potcar_strings = re.compile(r'\n?(\s*.*?End of Dataset\n)', re.S).findall(potcar_fo.read())

        for potcar_contents in potcar_strings:
            potcars.append(PotcarIo.from_(potcar_contents))
        return potcars

    @property
    def potcars(self) -> list[PotcarIo]:
        return self._potcars

    @classmethod
    def from_structure(cls, structure: orm.StructureData, potentials_dict: dict[str, Any]) -> MultiPotcarIo:
        """Create a MultiPotcarIo from an AiiDA `StructureData` object and a dictionary with a
        potential for each kind in the structure."""
        symbol_order = cls.potentials_order(structure)
        return cls(potcars=[potentials_dict[symbol] for symbol in symbol_order])

    def get_potentials_dict(self, structure: orm.StructureData) -> dict[str, Any]:
        """
        Get a dictionary {kind_name: PotcarData} that would fit the structure.

        If the PotcarData contained in MultiPotcarIo do not match the structure, an exception is raised.
        """
        structure_elements = structure.get_symbols_set()
        if structure_elements != self.element_symbols:
            raise ValueError('structure elements do not match POTCAR elements')
        if len(structure.get_kind_names()) != len(structure_elements):
            raise ValueError('structure has more kind names than elements')

        element_potcars = {potcario.node.element: potcario.node for potcario in self.potcars}
        return {kind.name: element_potcars[kind.symbol] for kind in structure.kinds}

    @property
    def element_symbols(self) -> set[str]:
        return {potcario.node.element for potcario in self.potcars}

    @classmethod
    def potentials_order(cls, structure: orm.StructureData) -> list[str]:
        return [kind[0] for kind in cls.count_kinds(structure)]

    @classmethod
    def count_kinds(cls, structure: orm.StructureData) -> list[tuple[str, int]]:
        """Count consecutive kinds that compose the different sites.

        :param structure: Structure containing sites and kinds
        :type structure: object
        :returns: List of tuples with kind names and counts
        :rtype: list
        """
        kind_name_order = [site.kind_name for site in structure.sites]
        groups = groupby(kind_name_order)
        counts = [(label, sum(1 for _ in group)) for label, group in groups]
        return counts



from time import sleep

import numpy as np
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import submit
from aiida.manage.configuration import load_profile
from aiida.orm import (
    Bool,
    Code,
    Float,
    Group,
    Int,
    QueryBuilder,
    Str,
    WorkChainNode,
    load_group,
)
from aiida.plugins import DataFactory, WorkflowFactory

load_profile()


def get_structure_SiC():
    """Set up SiC wurtzite cell

    wurtzite-type SiC
      1.0000000000
      3.0920000000   0.0000000000   0.0000000000
     -1.5460000000   2.6777505485   0.0000000000
      0.0000000000   0.0000000000   5.0730000000
    Si    C
        2     2
    Direct
      0.3333333333   0.6666666667   0.0000000000
      0.6666666667   0.3333333333   0.5000000000
      0.3333333333   0.6666666667   0.3758220000
      0.6666666667   0.3333333333   0.8758220000

    """

    StructureData = DataFactory('structure')
    a = 3.092
    c = 5.073
    lattice = [[a, 0, 0], [-a / 2, a / 2 * np.sqrt(3), 0], [0, 0, c]]
    structure = StructureData(cell=lattice)
    for pos_direct, symbol in zip(
        (
            [1.0 / 3, 2.0 / 3, 0],
            [2.0 / 3, 1.0 / 3, 0.5],
            [1.0 / 3, 2.0 / 3, 0.375822],
            [2.0 / 3, 1.0 / 3, 0.875822],
        ),
        ('Si', 'Si', 'C', 'C'),
    ):
        pos_cartesian = np.dot(pos_direct, lattice)
        structure.append_atom(position=pos_cartesian, symbols=symbol)
    return structure


def launch_aiida_relax_shape(structure, code_string, options, label):
    """Launch a relaxation of shape, but not volume."""

    Dict = DataFactory('dict')
    KpointsData = DataFactory('array.kpoints')

    # Set INCAR flags
    base_incar_dict = {
        'incar': {
            'PREC': 'Accurate',
            'EDIFF': 1e-8,
            'NELMIN': 5,
            'NELM': 100,
            'ENCUT': 500,
            'IALGO': 38,
            'ISMEAR': 0,
            'SIGMA': 0.01,
            'LREAL': False,
            'LCHARG': False,
            'LWAVE': False,
        }
    }

    # Set code, potentials to use and options
    base_config = {
        'code_string': code_string,
        'potential_family': 'PBE',
        'potential_mapping': {'Si': 'Si', 'C': 'C'},
        'options': options,
    }

    # Make sure some things are parsed
    base_parser_settings = {
        'add_energies': True,
        'add_forces': True,
        'add_stress': True,
    }

    # Set the provided code string
    code = Code.get_from_string(base_config['code_string'])

    # Which workchain to use
    workchain = WorkflowFactory('vasp.relax')

    # Construct the builder
    builder = workchain.get_builder()
    builder.code = code
    builder.parameters = Dict(dict=base_incar_dict)
    builder.structure = structure
    builder.settings = Dict(dict={'parser_settings': base_parser_settings})
    builder.potential_family = Str(base_config['potential_family'])
    builder.potential_mapping = Dict(dict=base_config['potential_mapping'])
    kpoints = KpointsData()
    kpoints.set_kpoints_mesh([6, 6, 4], offset=[0, 0, 0.5])
    builder.kpoints = kpoints
    builder.options = Dict(dict=base_config['options'])
    builder.metadata.label = label
    builder.metadata.description = label
    builder.clean_workdir = Bool(False)
    relax = AttributeDict()
    relax.perform = Bool(True)
    relax.force_cutoff = Float(1e-5)
    relax.positions = Bool(True)
    relax.shape = Bool(True)
    relax.volume = Bool(False)
    builder.relax = relax
    builder.verbose = Bool(True)

    # Submit calc
    node = submit(builder)
    return node


def launch_aiida_full_relax(structure, code_string, options, label):
    """Launch a relaxation of everything."""
    Dict = DataFactory('dict')
    KpointsData = DataFactory('array.kpoints')

    # Set INCAR flags
    base_incar_dict = {
        'incar': {
            'PREC': 'Accurate',
            'EDIFF': 1e-8,
            'NELMIN': 5,
            'NELM': 100,
            'ENCUT': 500,
            'IALGO': 38,
            'ISMEAR': 0,
            'SIGMA': 0.01,
            'LREAL': False,
            'LCHARG': False,
            'LWAVE': False,
        }
    }

    # Set code, potentials to use and options
    base_config = {
        'code_string': code_string,
        'potential_family': 'PBE.54',
        'potential_mapping': {'Si': 'Si', 'C': 'C'},
        'options': options,
    }

    # Make sure some things are parsed
    base_parser_settings = {
        'add_energies': True,
        'add_forces': True,
        'add_stress': True,
    }

    # Set the provided code string
    code = Code.get_from_string(base_config['code_string'])

    # Which workchain to use
    workchain = WorkflowFactory('vasp.relax')

    # Construct the builder
    builder = workchain.get_builder()
    builder.code = code
    builder.parameters = Dict(dict=base_incar_dict)
    builder.structure = structure
    builder.settings = Dict(dict={'parser_settings': base_parser_settings})
    builder.potential_family = Str(base_config['potential_family'])
    builder.potential_mapping = Dict(dict=base_config['potential_mapping'])
    kpoints = KpointsData()
    kpoints.set_kpoints_mesh([6, 6, 4], offset=[0, 0, 0.5])
    builder.kpoints = kpoints
    builder.options = Dict(dict=base_config['options'])
    builder.metadata.label = label
    builder.metadata.description = label
    builder.clean_workdir = Bool(False)
    relax = AttributeDict()
    relax.perform = Bool(True)
    relax.force_cutoff = Float(1e-5)
    relax.positions = Bool(True)
    relax.shape = Bool(True)
    relax.volume = Bool(True)
    relax.steps = Int(100)
    builder.relax = relax
    builder.verbose = Bool(True)

    # Submit the calc
    node = submit(builder)
    return node


def main(code_string, options, group_name, sleep_seconds=30):
    group = load_group(label=group_name)
    if not group.is_empty:
        # Make sure we do not already have members in the group
        print('The given group already have nodes. Please remove them or give a new group name.')
        exit(1)

    # Initial relaxation
    structure = get_structure_SiC()
    node_relax = launch_aiida_full_relax(structure, code_string, options, 'SiC VASP calc to relax volume')
    group.add_nodes(node_relax)
    print(f'Relaxing starting cell: {structure.cell}')

    while True:
        if node_relax.is_terminated:
            break
        print('Waiting for relaxation of starting cell to be done.')
        sleep(sleep_seconds)

    if node_relax.is_finished_ok:
        for strain, label in zip((0.99, 1.01), ('reduced', 'increased')):
            structure_clone = node_relax.outputs.relax.structure.clone()
            ase_structure = structure_clone.get_ase()
            # Make sure we also update atomic positions after scaling the cell
            ase_structure.set_cell(ase_structure.cell * strain ** (1.0 / 3.0), scale_atoms=True)
            StructureData = DataFactory('structure')
            structure = StructureData(ase=ase_structure)
            node = launch_aiida_relax_shape(
                structure,
                code_string,
                options,
                f'SiC VASP relax shape at {label} volume ({strain:f})',
            )
            group.add_nodes(node)
            print(f'Relaxation positions of scaled cell: {ase_structure.cell}')
    else:
        print('Relaxation failed.')
        exit(1)

    # Now make sure we wait until all calcs have been terminated
    while True:
        sleep(sleep_seconds)
        print('Waiting for all remaining relaxations to be done.')
        node_terminated = []
        for node in group.nodes:
            node_terminated.append(node.is_terminated)
        if all(node_terminated):
            break

    # Do one final check to make sure all calcs have a zero exit status
    for node in group.nodes:
        if not node.is_finished_ok:
            print(f'Node with id: {node.pk} has an exit status: {node.exit_status} and exit message: {node.exit_message}')
            exit(1)


def calc_bulk_modulus(group_name):
    stresses = []
    volumes = []
    for label in ('reduced', 'increased'):
        qb = QueryBuilder()
        qb.append(Group, filters={'label': group_name}, tag='group')
        qb.append(
            WorkChainNode,
            with_group='group',
            filters={'label': {'ilike': '%' + label + '%'}},
        )
        node = qb.first()[0]
        stresses.append(np.trace(node.outputs.stress.get_array('final')) / 3)
        volumes.append(np.linalg.det(node.inputs.structure.cell))

    d_s = stresses[1] - stresses[0]
    d_v = volumes[1] - volumes[0]
    v0 = (volumes[0] + volumes[1]) / 2.0
    bulk_modulus = -d_s / d_v * v0

    print(f'Bulk modulus: {bulk_modulus / 10.0} GPa')


if __name__ == '__main__':
    # Code_string is chosen from output of the list given by 'verdi code list'
    code_string = 'vasp@mycluster'

    # Set the options
    options = {
        'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 8},
        'account': '',
        'qos': '',
        'max_memory_kb': 2000000,
        'max_wallclock_seconds': 1800,
    }

    # Here we assume the group "Bulk_modulus_SiC_test" has already been created, but are empty
    group_name = 'Bulk_modulus_SiC_test'

    # Perform necessary calculations and put them in the created group
    main(code_string, options, group_name)

    # Calculate the bulk modulus by querying the group for the necessary input
    calc_bulk_modulus(group_name)



"""
Utility functions for running NEB calculations
"""

from __future__ import annotations

import numpy as np
from aiida import orm
from aiida.engine import calcfunction
from aiida.orm import StructureData
from ase.neb import NEB


@calcfunction
def neb_interpolate(
    init_structure: StructureData, final_structure: StructureData, nimages: orm.Int
) -> dict[str, StructureData]:
    """
    Interpolate NEB frames using the starting and the final structures

    Get around the PBC wrapping problem by calculating the MIC displacements
    from the initial to the final structure

    The initial structure is not changed, while the final structure is
    modified to be consistent with the initial structure in terms of absolute
    displacements, i.e. the final structure is *unwrapped*.
    """

    ainit = init_structure.get_ase()
    afinal = final_structure.get_ase()
    disps = []

    # Find distances
    acombined = ainit.copy()
    acombined.extend(afinal)
    # Get piece-wise MIC distances
    for i in range(len(ainit)):
        dist = acombined.get_distance(i, i + len(ainit), vector=True, mic=True)
        disps.append(dist.tolist())
    disps = np.asarray(disps)
    afinal = ainit.copy()

    # Displace the atoms according to MIC distances
    afinal.positions += disps
    neb = NEB([ainit.copy() for i in range(int(nimages) + 1)] + [afinal.copy()])
    neb.interpolate()
    out_init = StructureData(ase=neb.images[0])
    out_init.label = init_structure.label + ' INIT'
    out_final = StructureData(ase=neb.images[-1])
    out_final.label = init_structure.label + ' FINAL'

    outputs = {'image_init': out_init}
    for i, out in enumerate(neb.images[1:-1]):
        outputs[f'image_{i + 1:02d}'] = StructureData(ase=out)
        outputs[f'image_{i + 1:02d}'].label = init_structure.label + f' FRAME {i + 1:02d}'
    outputs['image_final'] = out_final
    return outputs


@calcfunction
def fix_atom_order(reference: StructureData, to_fix: StructureData) -> StructureData:
    """
    Fix atom order by finding NN distances between two frames. This resolves
    the issue where two closely matching structures having different atomic orders.
    Note that the two frames must be close enough for this to work
    """

    aref = reference.get_ase()
    afix = to_fix.get_ase()

    # Index of the reference atom in the second structure
    new_indices = np.zeros(len(aref), dtype=int)

    # Find distances
    acombined = aref.copy()
    acombined.extend(afix)
    # Get piece-wise MIC distances
    for i in range(len(aref)):
        dists = []
        for j in range(len(aref)):
            dist = acombined.get_distance(i, j + len(aref), mic=True)
            dists.append(dist)
        min_idx = np.argmin(dists)
        min_dist = min(dists)
        if min_dist > 0.5:
            print(f'Large displacement found - moving atom {j} to {i} - please check if this is correct!')
        new_indices[i] = min_idx

    afixed = afix[new_indices]
    fixed_structure = StructureData(ase=afixed)
    fixed_structure.label = to_fix.label + ' UPDATED ORDER'
    return fixed_structure



"""
Call script to calculate the total energies for different volumes of the silicon structure.

This particular call script calls the workchain that handles the execution of the
calculation of each structure. In this script we only supply a dictionary of different
structures as an input to the workchain, in addition to the standard inputs required by
the AiiDA-VASP workchain stack.
"""
# pylint: disable=too-many-arguments
import numpy as np
from aiida import load_profile
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import submit
from aiida.orm import Bool, Code, Str
from aiida.plugins import DataFactory
from eos_parallel import EosParallelWorkChain

load_profile()


def get_structure(alat):
    """
    Set up Si primitive cell

    fcc Si:
       alat
       0.5000000000000000    0.5000000000000000    0.0000000000000000
       0.0000000000000000    0.5000000000000000    0.5000000000000000
       0.5000000000000000    0.0000000000000000    0.5000000000000000
    Si
       1
    Cartesian
    0.0000000000000000  0.0000000000000000  0.0000000000000000

    """

    structure_data = DataFactory('structure')
    lattice = np.array([[0.5, 0.5, 0], [0, 0.5, 0.5], [0.5, 0, 0.5]]) * alat
    structure = structure_data(cell=lattice)
    for pos_direct in [[0.0, 0.0, 0.0]]:
        pos_cartesian = np.dot(pos_direct, lattice)
        structure.append_atom(position=pos_cartesian, symbols='Si')
        structure.label = f"silicon_at_{str(alat).replace('.', '_')}"
    return structure


def get_structures(lattice_constants):
    """Build a dictionary of structures using different lattice constants."""
    structures = {}
    for lattice_constant in lattice_constants:
        # use the lattice constant as a key
        structures[f"silicon_at_{str(lattice_constant).replace('.', '_')}"] = get_structure(lattice_constant)
    return structures


def main(code_string, incar, kmesh, structures, potential_family, potential_mapping, options):
    """Main method to setup the calculation."""

    # First, we need to fetch the AiiDA datatypes which will
    # house the inputs to our calculation
    dict_data = DataFactory('dict')
    kpoints_data = DataFactory('array.kpoints')

    # Then, we set the workchain you would like to call
    workchain = EosParallelWorkChain

    # And finally, we declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()

    # Set inputs for the following WorkChain execution
    # Code
    inputs.code = Code.get_from_string(code_string)
    # Structures
    inputs.structures = structures
    # k-points grid density
    kpoints = kpoints_data()
    kpoints.set_kpoints_mesh(kmesh)
    inputs.kpoints = kpoints
    # Parameters
    inputs.parameters = dict_data(dict=incar)
    # Potential family and their mapping between element and potential type to use
    inputs.potential_family = Str(potential_family)
    inputs.potential_mapping = dict_data(dict=potential_mapping)
    # Options
    inputs.options = dict_data(dict=options)
    # Settings
    inputs.settings = dict_data(dict=settings)
    # Workchain related inputs, in this case, give more explicit output to report
    inputs.verbose = Bool(True)
    # Submit the workchain with the set inputs
    submit(workchain, **inputs)


if __name__ == '__main__':
    # Code_string is chosen among the list given by 'verdi code list'
    CODE_STRING = 'vasp@mycluster'

    # INCAR equivalent
    # Set input parameters
    INCAR = {'incar': {'istart': 0, 'icharg': 2, 'encut': 240, 'ismear': 0, 'sigma': 0.1}}

    # KPOINTS equivalent
    # Set kpoint mesh
    KMESH = [11, 11, 11]

    # POTCAR equivalent
    # Potential_family is chosen among the list given by
    # 'verdi data vasp.potcar listfamilies'
    POTENTIAL_FAMILY = 'PBE.54'
    # The potential mapping selects which potential to use, here we use the standard
    # for silicon, this could for instance be {'Si': 'Si_GW'} to use the GW ready
    # potential instead
    POTENTIAL_MAPPING = {'Si': 'Si'}

    # jobfile equivalent
    # In options, we typically set scheduler options.
    # See https://aiida.readthedocs.io/projects/aiida-core/en/latest/scheduler/index.html
    # AttributeDict is just a special dictionary with the extra benefit that
    # you can set and get the key contents with mydict.mykey, instead of mydict['mykey']
    OPTIONS = AttributeDict()
    OPTIONS.account = ''
    OPTIONS.qos = ''
    OPTIONS.resources = {'num_machines': 1, 'num_mpiprocs_per_machine': 8}
    OPTIONS.queue_name = ''
    OPTIONS.max_wallclock_seconds = 3600
    OPTIONS.max_memory_kb = 2000000

    # POSCAR equivalent
    # Set the silicon structure
    LATTICE_CONSTANTS = [3.5, 3.6, 3.7, 3.8, 3.9, 4.0, 4.1, 4.2, 4.3]
    STRUCTURES = get_structures(LATTICE_CONSTANTS)

    main(
        CODE_STRING,
        INCAR,
        KMESH,
        STRUCTURES,
        POTENTIAL_FAMILY,
        POTENTIAL_MAPPING,
        OPTIONS,
    )
